import gzip
import logging

import os
import re
import tempfile
from io import StringIO
from typing import List, Dict
import getpass
import csv

import numpy as np
import pandas as pd
import pandasdmx
import requests
import requests_cache

from nexinfosys import get_global_configuration_variable
from nexinfosys.common.helper import create_dictionary, import_names, Memoize2, translate_case
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, filter_dataset_into_dataframe
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset, Dimension, CodeList, CodeImmutable


def create_estat_request():
    # EuroStat datasets
    if get_global_configuration_variable('CACHE_FILE_LOCATION'):
        cache_name = get_global_configuration_variable('CACHE_FILE_LOCATION')
        logging.debug(f"USER: {getpass.getuser()}, creating cache for Eurostat requests")
        if not os.path.isdir(cache_name):
            os.makedirs(cache_name)
    else:
        cache_name = tempfile.gettempdir() + "/sdmx_datasets_cache"
    r = pandasdmx.Request("ESTAT", cache={"backend": "sqlite", "include_get_headers": True, "cache_name": cache_name})
    r.timeout = 180
    return r


estat = None


class Eurostat(IDataSourceManager):
    def __init__(self):
        global estat
        if not estat:
            estat = create_estat_request()
        d = {"backend": "sqlite", "include_get_headers": True, "cache_name": tempfile.gettempdir() + "/eurostat_bulk_datasets"}
        requests_cache.install_cache(**d)

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "Eurostat"
        src.description = "Eurostat is the statistical office of the European Union"
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = ""
        db.description = "Eurostat provides all Datasets in a single database"
        return [db]

    @Memoize2
    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """
        import xmltodict
        lst = []
        # Make a table of datasets, containing three columns: ID, description, URN
        # List of datasets
        xml = requests.get("http://ec.europa.eu/eurostat/SDMX/diss-web/rest/dataflow/ESTAT/all/latest")
        t = xml.content.decode("utf-8")
        j = xmltodict.parse(t)
        for k in j["mes:Structure"]["mes:Structures"]["str:Dataflows"]["str:Dataflow"]:
            for n in k["com:Name"]:
                if n["@xml:lang"] == "en":
                    desc = n["#text"]
                    break
            if k["@id"][:3] != "DS-":  # or k["@id"] in ("DS-066341", "DS-066342", "DS-043408", "DS-043409"):
                # dsd_id = k["str:Structure"]["Ref"]["@id"]
                lst.append((k["@id"], desc, k["@urn"]))

            # print(dsd_id + "; " + desc + "; " + k["@id"] + "; " + k["@urn"])
        return lst

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """
        refs = dict(references='all')
        dsd_response = estat.datastructure("DSD_" + dataset, params=refs)
        dsd = dsd_response.datastructure["DSD_" + dataset]
        metadata = dsd_response.write()
        # SDMXConcept = collections.namedtuple('Concept', 'type name istime description code_list')
        # DataSource <- Database <- DATASET <- Dimension(s) (including Measures) <- CodeList
        #                                      |
        #                                      v
        #                                      Concept <- CodeList  (NOT CONSIDERED NOW)
        ds = Dataset()
        ds.code = dataset
        ds.description = None  # How to get description?
        ds.attributes = {}  # Dataset level attributes? (encode them using a dictionary)
        ds.metadata = None  # Metadata for the dataset SDMX (flow, date of production, etc.)
        ds.database = database  # Reference to containing database

        dims = {}

        for d in dsd.dimensions:
            istime = str(dsd.dimensions.get(d)).split("|")[0].strip() == "TimeDimension"
            dd = Dimension()
            dd.code = d
            dd.description = None
            dd.attributes = None
            dd.is_time = istime
            dd.is_measure = False
            dd.dataset = ds
            dims[d] = dd
        for m in dsd.measures:
            dd = Dimension()
            dd.code = m
            dd.description = None
            dd.attributes = None
            dd.is_time = False
            dd.is_measure = True
            dd.dataset = ds
            dims[m] = dd
        for a in dsd.attributes:
            ds.attributes[a] = None  # TODO Get the value
        for l in metadata.codelist.index.levels[0]:
            first = True
            # Read code lists
            cl = create_dictionary()
            for m, v in list(zip(metadata.codelist.loc[l].index, metadata.codelist.loc[l]["name"])):
                if not first:
                    cl[m] = v.replace("\n", ";")
                else:
                    first = False
            # Attach it to the Dimension or Measure
            if metadata.codelist.loc[l]["dim_or_attr"][0] == "D":
                # Build Code List from dictionary
                dims[l].code_list = CodeList.construct(l, None, [""], [CodeImmutable(k, cl[k], "", []) for k in cl])

        return ds

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False) -> str:
        """
        Download a file (general purpose, not only for Eurostat datasets)

        :param url:
        :param local_filename:
        :param update:
        :return: String with full file name
        """
        url = "http://ec.europa.eu/eurostat/estat-navtree-portlet-prod/BulkDownloadListing?downfile=data%2F" + dataset + ".tsv.gz"
        zip_name = tempfile.gettempdir() + "/" + dataset + '.tsv.gz'
        if os.path.isfile(zip_name):
            if not update:
                return zip_name
            else:
                os.remove(zip_name)

        r = requests.get(url, stream=True)
        # http://stackoverflow.com/questions/15352668/download-and-decompress-gzipped-file-in-memory
        with open(zip_name, 'wb') as f:
            for chunk in r.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)

        return zip_name

    def get_dataset_filtered(self, dataset: str, dataset_params: dict) -> Dataset:
        """ This method has to consider the last dataset download, to re"""

        def multi_replace(text, rep):
            rep = dict((re.escape(k), v) for k, v in rep.items())
            pattern = re.compile("|".join(rep.keys()))
            return pattern.sub(lambda m: rep[re.escape(m.group(0))], text)

        # Read Eurostat dataset structure
        ds = self.get_dataset_structure(None, dataset)

        # Read full Eurostat dataset into a Dataframe
        dataframe_fn = tempfile.gettempdir() + "/" + dataset + ".bin2"
        df = None
        if os.path.isfile(dataframe_fn):
            df = pd.read_parquet(dataframe_fn)
            # df = pd.read_msgpack(dataframe_fn)

        if df is None:
            zip_name = self.etl_dataset(dataset, update=False)
            new_method = False
            if new_method:
                # TODO Obtain a Dataframe with pairs of columns for observations
                with gzip.open(zip_name, "rb") as gz:
                    st = None
                    fc = StringIO(st)
                    # Read header
                    csv_reader = csv.reader(fc)
                    header = next(csv_reader)
                    fc.seek(0)
                    # Obtain real header: each period column is added a
                    # Parse it

            else:
                with gzip.open(zip_name, "rb") as gz:
                    # Read file and
                    # Remove status flags (documented at http://ec.europa.eu/eurostat/data/database/information)
                    #
                    # TODO ISOLATE STATUS FLAGS INTO ANOTHER COLUMN
                    # TODO ":" -> "NaN\t",
                    # TODO ([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)(.*) -> (1)\t(2)
                    pattern = re.compile("(\\:)|( [b-fnpruzscde]+)")
                    st = pattern.sub(lambda m: "NaN" if m.group(0) == ":" else "", gz.read().decode("utf-8"))

                    # st = multi_replace(gz.read().decode("utf-8"),
                    #                    {":": "NaN", " p": "", " e": "", " f": "", " n": "", " c": "", " u": "",
                    #                     " z": "", " r": "", " b": "", " d": ""})
                    fc = StringIO(st)
                    # fc = StringIO(gz.read().decode("utf-8").replace(" p\t", "\t").replace(":", "NaN"))
                os.remove(zip_name)
                # Remove ":" -> NaN
                # Remove " p" -> ""
                df = pd.read_csv(fc, sep="\t")

                def split_codes(all_codes):  # Split and strip
                    return [s.strip() for s in all_codes.split(",")]

                original_column = df.columns[0]
                new_cols = [s.strip() for s in original_column.split(",")]
                new_cols[-1] = new_cols[-1][:new_cols[-1].find("\\")]
                temp = list(zip(*df[original_column].map(split_codes)))
                del df[original_column]
                df.columns = [c.strip() for c in df.columns]
                # Convert to numeric
                for cn in df.columns:
                    try:
                        df[cn] = df[cn].astype(np.float)
                    except ValueError:
                        print("BORRAME - Error conversion columna "+cn+", probablemente queda algun flag sin eliminar")
                    # df[cn] = df[cn].apply(lambda x: pd.to_numeric(x, errors='coerce'))
                # Add index columns
                for i, c in enumerate(new_cols):
                    df[c] = temp[i]
                # set index on the dimension columns
                df.set_index(new_cols, inplace=True)
                # Save df
                df.to_parquet(dataframe_fn)
                # df.to_msgpack(dataframe_fn)

        # Change dataframe index names to match the case of the names in the metadata
        # metadata_names_dict = {dim.code.lower(): dim.code for dim in ds.dimensions}
        # dataframe_new_names = [metadata_names_dict.get(name.lower(), name) for name in df.index.names]
        dataframe_new_names = translate_case(df.index.names, [dim.code for dim in ds.dimensions])
        df.index.names = dataframe_new_names

        # Filter it using generic Pandas filtering capabilities
        if dataset_params:
            ds.data = filter_dataset_into_dataframe(df, dataset_params, dataset, eurostat_postprocessing=True)
        else:
            ds.data = df

        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass


def multi_replace(text, rep):
    import re
    rep = dict((re.escape(k), v) for k, v in rep.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], text)


if __name__ == '__main__':
    e = Eurostat()
    e.get_dataset_filtered("sbs_na_ind_r2", None)



import io

import networkx as nx

from nexinfosys.models.musiasem_concepts import FactorsRelationScaleObservation
from nexinfosys.solving import *
from nexinfosys import ureg
from nexinfosys.command_generators.parser_ast_evaluators import ast_to_string
from nexinfosys.models.musiasem_concepts_helper import find_or_create_observable, find_quantitative_observations
from nexinfosys.serialization import deserialize_state
from nexinfosys.command_generators import parser_field_parsers

"""
IN DEVELOPMENT. NOT READY FOR INTEGRATION INTO THE DOCKER CONTAINER

This module will contain functions to compose graphs for different purposes

* For Internal/External Processes capable of exploiting the graph topology and associated attributes

* For deduction of Factors:
  - Nodes: Factors
  - Edges: flow relations, relations between FactorTypes (hierarchy, expressions in hierarchies, linear transforms)


* For summary visualization of flow relation between Processors 
  - Nodes: Processors
  - Edges: Flows between factors, several flows between the same pair of processor are summarized into one

* Visualization of graph with hierarchy of processors 
  - Nodes: Processors
  - Edges: part-of relations
* Visualization of Processors and their factors
  - Nodes: Processors and Factors
  - Edges: Processors to Factors, flows Factor to Factor

Allow filtering:
  - By Observer
  - By other criteria 


>> Si hay expresiones no lineales, np hay una manera directa de expresar en el grafo <<

Mapa

Parámetros - Variables SymPy
Factores - Variables SymPy

ECUACIONES DIRECTAMENTE

Factores -> Variables
Parámetros -> Inputs

Observaciones extensivas -> Valores para las variables
Observaciones intensivas -> ECUACIÓN: F2 = (V)*F1
Flujo entre factores: padre <- hijos: F = sum(Fhijos)
Flujo entre factores: SPLIT: F1 = W1*F, F2 = W2*F ó F = F1/W1 + F2/W2 + ...
Flujo entre factores: JOIN : F  = W1*F1 + W2*F2 + ...
Ambos son redundantes

F1  G1
F2  G2
G1 = W1*F1 + W2*F2
G2 = (1-W1)*F1 + (1-W2)*F2  
"""


def construct_flow_graph_2(state: State, query: IQueryObjects, filt: Union[str, dict], format: str="visjs"):
    """
    Prepare a graph from which conclusions about factors can be extracted

    Example:
        1) Obtain "s", the serialized state from Redis or from a test file
        2) state = deserialize_state(s)
        3) query = BasicQuery(state) # Create a Query and execute a query
        4) construct_solve_graph(state, query, None)

    :param state: State
    :param query: A IQueryObjects instance (which has been already injected the state)
    :param filt: A filter to be passed to the query instance
    :param format: VisJS, GML, ...
    :return:
    """
    include_processors = False  # For completeness (not clarity...), include processors nodes, as a way to visualize grouped factors
    will_write = True  # For debugging purposes, affects how the properties attached to nodes and edges are elaborated
    expand_factors_graph = False  # Expand transformation between FactorTypes into instances of Factors

    # Format for different node types
    stated_factor_no_observation = dict(graphics={'fill': "#999900"})  # Golden
    stated_factor_some_observation = dict(graphics={'fill': "#ffff00"})  # Yellow
    qq_attached_to_factor = dict(graphics={'fill': "#eeee00", "type": "ellipse"})  # Less bright Yellow
    non_stated_factor = dict(graphics={'fill': "#999999"})
    a_processor = dict(graphics={"type": "hexagon", "color": "#aa2211"})

    # Format for different edge types
    edge_from_factor_type = dict(graphics={"fill": "#ff0000", "width": 1, "targetArrow": "standard"})
    edge_processor_to_factor = dict(graphics={"fill": "#ff00ff", "width": 3, "targetArrow": "standard"})
    edge_factors_flow = dict(graphics={"fill": "#000000", "width": 5, "targetArrow": "standard"})
    edge_factors_scale = dict(graphics={"fill": "#333333", "width": 3, "targetArrow": "standard"})
    edge_factors_relative_to = dict(graphics={"fill": "#00ffff", "width": 3, "targetArrow": "standard"})
    edge_factor_value = dict(graphics={"fill": "#aaaaaa", "width": 1, "targetArrow": "standard"})

    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    # Obtain the information needed to elaborate the graph
    objs = query.execute([Processor, Factor, FactorType,
                          FactorTypesRelationUnidirectionalLinearTransformObservation,
                          FactorsRelationScaleObservation,
                          FactorsRelationDirectedFlowObservation
                          ],
                         filt
                         )

    # 1) Graphical Representation: BOX -- BOX
    #
    # 2) Internal (not for end-users), pseudo-code:
    #
    # Processor1 <- Factor1 -> FactorType0
    # Processor2 <- Factor2 -> FactorType0
    # Processor3 <- Factor3 -> FactorType1
    # Processor3 <- Factor4 -> FactorType0
    # Factor1 <- FactorsRelationDirectedFlowObservation(0.4) -> Factor2
    # Factor1 <- FactorsRelationDirectedFlowObservation(0.6) -> Factor4
    # Factor1 <- FactorQuantitativeObservation(5.3 m²)
    # FactorType0 -> FactorTypesRelationUnidirectionalLinearTransformObservation(ctx) -> FactorType1
    # Factor4 -> w1 -> Factor3
    # Factor5 -> w2 -> Factor3
    #

    # Index quantitative observations.
    # Also, mark Factors having QQs (later this will serve to color differently these nodes)
    qqs = {}
    qq_cont = 0
    factors_with_some_observation = set()
    for o in find_quantitative_observations(glb_idx):
        # Index quantitative observations.
        if "relative_to" in o.attributes and o.attributes["relative_to"]:
            continue  # Do not index intensive quantities, because they are translated as edges in the graph
        if o.factor in qqs:
            lst = qqs[o.factor]
        else:
            lst = []
            qqs[o.factor] = lst
        lst.append(o)
        # Mark Factors having QQs (later this will serve to color differently these nodes)
        factors_with_some_observation.add(o.factor)

    # ---- MAIN GRAPH: Factors and relations between them --------------------------------------------------------------
    the_node_names_set = set()

    # --   Nodes: "Factor"s passing the filter, and QQs associated to some of the Factors
    n = []
    e = []
    f_types = {}  # Contains a list of Factors for each FactorType
    p_factors = {}  # Contains a list of Factors per Processor7
    rel_to_observations = set()  # Set of FactorObservation having "relative_to" property defined
    factors = create_dictionary()  # Factor_ID -> Factor
    for f in objs[Factor]:
        f_id = get_factor_id(f, prd=glb_idx)
        factors[f_id] = f  # Dictionary Factor_ID -> Factor
        # f_types
        if f.taxon in f_types:
            lst = f_types[f.taxon]
        else:
            lst = []
            f_types[f.taxon] = lst
        lst.append(f)
        # p_factors
        if f.processor in p_factors:
            lst = p_factors[f.processor]
        else:
            lst = []
            p_factors[f.processor] = lst
        lst.append(f)

        # Add Node to graph
        the_node_names_set.add(f_id)
        if will_write:
            n.append((f_id, stated_factor_some_observation if f in factors_with_some_observation else stated_factor_no_observation))
            if f in qqs:
                for qq in qqs[f]:
                    if not ("relative_to" in qq.attributes and qq.attributes["relative_to"]):
                        # value = str(qq.value)  # str(qq_cont) + ": " + str(qq.value)
                        value_node_name = f_id + " " + str(qq.value)
                        n.append((value_node_name, qq_attached_to_factor))
                        e.append((value_node_name, f_id, {"w": "", "label": "", **edge_factor_value}))
                        qq_cont += 1
                    else:
                        rel_to_observations.add(qq)
        else:
            qqs2 = [qq for qq in qqs if not ("relative_to" in qq.attributes and qq.attributes["relative_to"])]
            d = dict(factor=factor_to_dict(f), observations=qqs[f_id] if f_id in qqs2 else [])
            n.append((f_id, d))

    # --   Edges
    # "Relative to" relation (internal to the Processor) -> Intensive to Extensive
    for o in rel_to_observations:
        if "relative_to" in o.attributes and o.attributes["relative_to"]:
            # Parse "defining_factor", it can be composed of the factor name AND the unit
            defining_factor = o.attributes["relative_to"]
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.factor_unit, defining_factor)
            factor_type = ast_to_string(ast["factor"])
            unit_name = ast["unparsed_unit"]
            ureg(unit_name)
            f_id = get_factor_id(o.factor, prd=glb_idx)
            # Check that "f_id" exists in the nodes list (using "factors")
            factors[f_id]
            # If "defining_factor" exists in the processor, ok. If not, create it.
            # Find factor_type in the processor
            factor_name = get_processor_id(o.factor.processor) + ":" + factor_type
            factors[factor_name]
            e.append((factor_name, f_id, {"w": o.value.expression, "label": o.value.expression, **edge_factors_relative_to}))

    # Directed Flows between Factors
    for df in objs[FactorsRelationDirectedFlowObservation]:
        sf = get_factor_id(df.source_factor, prd=glb_idx)
        tf = get_factor_id(df.target_factor, prd=glb_idx)
        # Check that both "sf" and "tf" exist in the nodes list (using "factors")
        factors[sf]
        factors[tf]
        weight = df.weight if df.weight else "1"
        e.append((sf, tf, {"w": weight, "label": weight, **edge_factors_flow}))

    # Scale Flows between Factors
    for df in objs[FactorsRelationScaleObservation]:
        sf = get_factor_id(df.origin, prd=glb_idx)
        tf = get_factor_id(df.destination, prd=glb_idx)
        # Check that both "sf" and "tf" exist in the nodes list (using "factors")
        factors[sf]
        factors[tf]
        weight = str(df.quantity) if df.quantity else "1"
        e.append((sf, tf, {"w": weight, "label": weight, **edge_factors_scale}))

    # TODO Consider Upscale relations
    # e.append((..., ..., {"w": upscale_weight, "label": upscale_weight, **edge_factors_upscale}))

    # -- Create the graph
    factors_graph = nx.DiGraph()
    factors_graph.add_nodes_from(n)
    factors_graph.add_edges_from(e)

    # nx.write_gml(factors_graph, "/home/rnebot/IntermediateGraph.gml")

    # ---- AUXILIARY GRAPH: FACTOR TYPES AND THEIR INTERRELATIONS ----
    n = []
    e = []
    # --   Nodes: "FactorType"s passing the filter
    for ft in objs[FactorType]:
        n.append((get_factor_type_id(ft), dict(factor_type=ft)))

    # --   Edges
    # Hierarchy and expressions stated in the hierarchy
    ft_in = {}  # Because FactorTypes cannot be both in hierarchy AND expression, marks if it has been specified one was, to raise an error if it is specified also the other way
    for ft in objs[FactorType]:
        ft_id = get_factor_type_id(ft)
        if ft.expression:
            if ft not in ft_in:
                # TODO Create one or more relations, from other FactorTypes (same Hierarchy) to this one
                # TODO The expression can only be a sum of FactorTypes (same Hierarchy)
                ft_in[ft] = "expression"
                # TODO Check that both "ft-id" and "..." exist in the nodes list (keep a temporary set)
                # weight = ...
                # e.append((ft_id, ..., {"w": weight, "label": weight, "origin": ft, "destination": ...}))

        if ft.parent:
            if ft.parent not in ft_in or (ft.parent in ft_in and ft_in[ft.parent] == "hierarchy"):
                # Create an edge from this FactorType
                ft_in[ft.parent] = "hierarchy"
                parent_ft_id = get_factor_type_id(ft.parent)
                # TODO Check that both "ft-id" and "parent_ft_id" exist in the nodes list (keep a temporary set)
                # Add the edge
                e.append((ft_id, parent_ft_id, {"w": "1", "origin": ft, "destination": ft.parent}))
            else:
                raise Exception("The FactorType '"+ft_id+"' has been specified by an expression, it cannot be parent.")
    # Linear transformations
    for f_rel in objs[FactorTypesRelationUnidirectionalLinearTransformObservation]:
        origin = get_factor_type_id(f_rel.origin)
        destination = get_factor_type_id(f_rel.destination)
        e.append((origin, destination, {"w": f_rel.weight, "label": f_rel.weight, "origin": f_rel.origin, "destination": f_rel.destination}))

    # ---- Create FACTOR TYPES graph ----

    factor_types_graph = nx.DiGraph()
    factor_types_graph.add_nodes_from(n)
    factor_types_graph.add_edges_from(e)

    # ---- EXPAND "FACTORS GRAPH" with "FACTOR TYPE" RELATIONS ----
    sg_list = []  # List of modified (augmented) subgraphs
    if expand_factors_graph:
        # The idea is: clone a FactorTypes subgraph if a Factor instances some of its member nodes
        # This cloning process can imply creating NEW Factors

        the_new_node_names_set = set()

        # Obtain weak components of the main graph. Each can be considered separately

        # for sg in nx.weakly_connected_component_subgraphs(factors_graph):  # For each subgraph
        #     print("--------------------------------")
        #     for n in sg.nodes():
        #         print(n)

        # ---- Weakly connected components of "factor_types_graph" ----
        factor_types_subgraphs = list(nx.weakly_connected_component_subgraphs(factor_types_graph))

        for sg in nx.weakly_connected_component_subgraphs(factors_graph):  # For each subgraph
            sg_list.append(sg)
            # Consider each Factor of the subgraph
            unprocessed_factors = set(sg.nodes())
            while unprocessed_factors:  # For each UNPROCESSED Factor
                tmp = unprocessed_factors.pop()  # Get next unprocessed "factor name"
                if tmp not in factors:  # QQ Observations are in the graph and not in "factors". The same with Processors
                    continue
                f_ = factors[tmp]  # Obtain Factor from "factor name"
                ft_id = get_factor_type_id(f_)  # Obtain FactorType name from Factor
                # Iterate through FactorTypes and check if the Factor appears
                for sg2 in factor_types_subgraphs:  # Each FactorTypes subgraph
                    if ft_id in sg2:  # If the current Factor is in the subgraph
                        if len(sg2.nodes()) > 1:  # If the FactorType subgraph has at least two nodes
                            # CLONE FACTOR TYPES SUBGRAPH
                            # Nodes. Create if not present already
                            n = []
                            e = []
                            for n2, attrs in sg2.nodes().items():  # Each node in the FactorTypes subgraph
                                ft_ = attrs["factor_type"]
                                f_id = get_factor_id(f_.processor, ft_, prd=glb_idx)
                                if f_id not in sg:  # If the FactorType is not
                                    # Create Factor, from processor and ft_ -> f_new
                                    _, _, f_new = find_or_create_observable(state, name=f_id, source="solver")
                                    factors[f_id] = f_new
                                    if f_id not in the_node_names_set:
                                        if will_write:
                                            n.append((f_id, non_stated_factor))
                                        else:
                                            d = dict(factor=factor_to_dict(f_new), observations=[])
                                            n.append((f_id, d))
                                    if f_id not in the_node_names_set:
                                        the_new_node_names_set.add(f_id)
                                    the_node_names_set.add(f_id)
                                else:
                                    unprocessed_factors.discard(f_id)
                            # Edges. Create relations between factors
                            for r2, w_ in sg2.edges().items():
                                # Find origin and destination nodes. Copy weight. Adapt weight? If it refers to a FactorType, instance it?
                                origin = get_factor_id(f_.processor, w_["origin"], prd=glb_idx)
                                destination = get_factor_id(f_.processor, w_["destination"], prd=glb_idx)
                                if origin in the_new_node_names_set or destination in the_new_node_names_set:
                                    graphics = edge_from_factor_type
                                else:
                                    graphics = {}
                                e.append((origin, destination, {"w": w_["w"], "label": w_["w"], **graphics}))
                            sg.add_nodes_from(n)
                            sg.add_edges_from(e)
                            break

        # for sg in sg_list:
        #     print("--------------------------------")
        #     for n in sg.nodes():
        #         print(n)

    # Recompose the original graph
    if sg_list:
        factors_graph = nx.compose_all(sg_list)
    else:
        pass
        ##factors_graph = nx.DiGraph()

    # ----
    # Add "Processor"s just as a way to visualize grouping of factors (they do not influence relations between factors)
    # -
    if include_processors:
        n = []
        e = []
        for p in objs[Processor]:
            p_id = get_processor_id(p)
            if will_write:
                n.append((p_id, a_processor))
            else:
                n.append((p_id, processor_to_dict(p)))
            # Edges between Processors and Factors
            for f in p_factors[p]:
                f_id = get_factor_id(f, prd=glb_idx)
                e.append((p_id, f_id, edge_processor_to_factor))
        factors_graph.add_nodes_from(n)
        factors_graph.add_edges_from(e)

    #
    # for ft in objs[FactorType]:
    #     if ft.parent:
    #         # Check which Factors are instances of this FactorType
    #         if ft in f_types:
    #             for f in f_types[ft]:
    #                 # Check if the processor contains the parent Factor
    #                 processor_factors = p_factors[f.processor]
    #                 if ft.parent not in processor_factors:
    #                     factor_data = (f.processor, ft)
    #                 else:
    #                     factor_data = None
    #                 create_factor = f in qqs  # If there is some Observation
    #                 create_factor = True # Force creation
    #
    #
    #         # Consider the creation of a relation
    #         # Consider also the creation of a new Factor (a new Node for now): if the child has some observation for sure (maybe a child of the child had an observation, so it is the same)
    #         ft_id =
    #     ft_id =

    # Plot graph to file
    # import matplotlib.pyplot as plt
    # ax = plt.subplot(111)
    # ax.set_title('Soslaires Graph', fontsize=10)
    # nx.draw(factors_graph, with_labels=True)
    # plt.savefig("/home/rnebot/Graph.png", format="PNG")

    # GML File
    # nx.write_gml(factors_graph, "/home/rnebot/Graph.gml")

    ret = None
    if format == "visjs":
        # Assign IDs to nodes. Change edges "from" and "to" accordingly
        ids_map = create_dictionary()
        id_count = 0
        for node in factors_graph.nodes(data=True):
            sid = str(id_count)
            node[1]["id"] = sid
            ids_map[node[0]] = sid
            id_count += 1

        vis_nodes = []
        vis_edges = []
        for node in factors_graph.nodes(data=True):
            d = dict(id=node[1]["id"], label=node[0])
            if "shape" in node[1]:
                # circle, ellipse, database, box, diamond, dot, square, triangle, triangleDown, text, star
                d["shape"] = node[1]["shape"]
            else:
                d["shape"] = "box"
            if "color" in node[1]:
                d["color"] = node[1]["color"]
            vis_nodes.append(d)
        for edge in factors_graph.edges(data=True):
            f = ids_map[edge[0]]
            t = ids_map[edge[1]]
            d = {"from": f, "to": t, "arrows": "to"}
            data = edge[2]
            if "w" in data:
                d["label"] = data["w"]
                d["font"] = {"align": "horizontal"}

            vis_edges.append(d)
        ret = {"nodes": vis_nodes, "edges": vis_edges}
    elif format == "gml":
        ret1 = io.BytesIO()
        nx.write_gml(factors_graph, ret1)
        ret = ret1.getvalue()
        ret1.close()

    return ret

# #########################################################################3

    # GEXF File
    # nx.write_gexf(factors_graph, "/home/rnebot/Graph.gexf")

    # Legend graph
    n = []
    e = []
    n.append(("Factor with Observation", stated_factor_some_observation))
    n.append(("Factor with No Observation", stated_factor_no_observation))
    if include_processors:
        n.append(("Processor", a_processor))
    n.append(("Factor from FactorType", non_stated_factor))
    n.append(("QQ Observation", qq_attached_to_factor))
    n.append(("QQ Intensive Observation", qq_attached_to_factor))

    e.append(("A Factor", "Another Factor", {"label": "Flow between Factors, attaching the weight", **edge_factors_flow}))
    e.append(("Factor #1", "Factor #2", {"label": "Relation from a FactorType", **edge_from_factor_type}))
    if include_processors:
        e.append(("Processor", "A Factor", {"label": "Link from Processor to Factor", **edge_processor_to_factor}))
    e.append(("A Factor", "Same Factor in another processor", {"label": "Upscale a Factor in two processors", **edge_factors_upscale}))
    e.append(("Factor with Observation", "QQ Intensive Observation", {"label": "Observation proportional to extensive value of factor same processor", **edge_factors_relative_to}))
    e.append(("QQ Observation", "A Factor", {"label": "A QQ Observation", **edge_factor_value}))
    factors_graph = nx.DiGraph()
    factors_graph.add_nodes_from(n)
    factors_graph.add_edges_from(e)
    # nx.write_gml(factors_graph, "/home/rnebot/LegendGraph.gml")


def construct_flow_graph(state: State, query: IQueryObjects, filt: Union[str, dict], format: str="visjs"):
    """
    Prepare a graph from which conclusions about factors can be extracted

    Example:
        1) Obtain "s", the serialized state from Redis or from a test file
        2) state = deserialize_state(s)
        3) query = BasicQuery(state) # Create a Query and execute a query
        4) construct_solve_graph(state, query, None)

    :param state: State
    :param query: A IQueryObjects instance (which has been already injected the state)
    :param filt: A filter to be passed to the query instance
    :return:
    """
    include_processors = False  # For clarity, include processors nodes, as a way to visualize grouped factors
    will_write = True  # For debugging purposes, affects how the properties attached to nodes and edges are elaborated
    # Format
    stated_factor_no_observation = dict(graphics={'fill': "#999900"})  # Golden
    stated_factor_some_observation = dict(graphics={'fill': "#ffff00"})  # Yellow
    qq_attached_to_factor = dict(graphics={'fill': "#eeee00", "type": "ellipse"})  # Less bright Yellow
    non_stated_factor = dict(graphics={'fill': "#999999"})
    a_processor = dict(graphics={"type": "hexagon", "color": "#aa2211"})

    edge_from_factor_type = dict(graphics={"fill": "#ff0000", "width": 1, "targetArrow": "standard"})
    edge_processor_to_factor = dict(graphics={"fill": "#ff00ff", "width": 3, "targetArrow": "standard"})
    edge_factors_flow = dict(graphics={"fill": "#000000", "width": 5, "targetArrow": "standard"})
    edge_factors_upscale = dict(graphics={"fill": "#333333", "width": 3, "targetArrow": "standard"})
    edge_factors_relative_to = dict(graphics={"fill": "#00ffff", "width": 3, "targetArrow": "standard"})
    edge_factor_value = dict(graphics={"fill": "#aaaaaa", "width": 1, "targetArrow": "standard"})

    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

    # Obtain the information needed to elaborate the graph
    objs = query.execute([Processor, Factor, FactorType,
                          FactorTypesRelationUnidirectionalLinearTransformObservation,
                          ProcessorsRelationPartOfObservation, ProcessorsRelationUpscaleObservation,
                          ProcessorsRelationUndirectedFlowObservation,
                          FactorsRelationDirectedFlowObservation
                          ],
                         filt
                         )

    # Processor <- ProcessorsRelationPartOfObservation -> Processor

    # 1) Graphical Representation: BOX -- BOX
    #
    # 2) Internal (not for end-users), pseudo-code:
    #
    # Processor1 <- Factor1 -> FactorType0
    # Processor2 <- Factor2 -> FactorType0
    # Processor3 <- Factor3 -> FactorType1
    # Processor3 <- Factor4 -> FactorType0
    # Factor1 <- FactorsRelationDirectedFlowObservation(0.4) -> Factor2
    # Factor1 <- FactorsRelationDirectedFlowObservation(0.6) -> Factor4
    # Factor1 <- FactorQuantitativeObservation(5.3 m²)
    # FactorType0 -> FactorTypesRelationUnidirectionalLinearTransformObservation(ctx) -> FactorType1
    # Factor4 -> w1 -> Factor3
    # Factor5 -> w2 -> Factor3
    #

    # EIM:
    # LETs "eco"
    # for interface in eco.interfaces:
    #     if
    #
    # EUM:

    # 3) Spreadsheet representation
    # ....

    # Index quantitative observations.
    # Also, mark Factors having QQs (later this will serve to color differently these nodes)
    qqs = {}
    qq_cont = 0
    factors_with_some_observation = set()
    for o in find_quantitative_observations(glb_idx):
        # Index quantitative observations.
        if "relative_to" in o.attributes and o.attributes["relative_to"]:
            continue  # Do not index intensive quantities, because they are translated as edges in the graph
        if o.factor in qqs:
            lst = qqs[o.factor]
        else:
            lst = []
            qqs[o.factor] = lst
        lst.append(o)
        # Mark Factors having QQs (later this will serve to color differently these nodes)
        factors_with_some_observation.add(o.factor)

    # ---- MAIN GRAPH: Factors and relations between them --------------------------------------------------------------
    the_node_names_set = set()

    # --   Nodes: "Factor"s passing the filter, and QQs associated to some of the Factors
    n = []
    e = []
    f_types = {}  # Contains a list of Factors for each FactorType
    p_factors = {}  # Contains a list of Factors per Processor
    factors = create_dictionary() # Factor_ID -> Factor
    for f in objs[Factor]:
        f_id = get_factor_id(f)
        factors[f_id] = f  # Dictionary Factor_ID -> Factor
        # f_types
        if f.taxon in f_types:
            lst = f_types[f.taxon]
        else:
            lst = []
            f_types[f.taxon] = lst
        lst.append(f)
        # p_factors
        if f.processor in p_factors:
            lst = p_factors[f.processor]
        else:
            lst = []
            p_factors[f.processor] = lst
        lst.append(f)

        # Node
        the_node_names_set.add(f_id)
        if will_write:
            n.append((f_id, stated_factor_some_observation if f in factors_with_some_observation else stated_factor_no_observation))
            if f in qqs:
                for qq in qqs[f]:
                    if not ("relative_to" in qq.attributes and qq.attributes["relative_to"]):
                        value = str(qq.value)  # str(qq_cont) + ": " + str(qq.value)
                        n.append((value, qq_attached_to_factor))
                        e.append((value, f_id, {"w": "", "label": "", **edge_factor_value}))
                        qq_cont += 1
        else:
            qqs2 = [qq for qq in qqs if not ("relative_to" in qq.attributes and qq.attributes["relative_to"])]
            d = dict(factor=factor_to_dict(f), observations=qqs[f_id] if f_id in qqs2 else [])
            n.append((f_id, d))

    # --   Edges
    # "Relative to" relation (internal to the Processor) -> Intensive to Extensive
    for o in objs[FactorQuantitativeObservation]:
        if "relative_to" in o.attributes and o.attributes["relative_to"]:
            # Parse "defining_factor", it can be composed of the factor name AND the unit
            defining_factor = o.attributes["relative_to"]
            ast = parser_field_parsers.string_to_ast(parser_field_parsers.factor_unit, defining_factor)
            factor_type = ast_to_string(ast["factor"])
            unit_name = ast["unparsed_unit"]
            ureg(unit_name)
            f_id = get_factor_id(o.factor)
            # Check that "f_id" exists in the nodes list (using "factors")
            factors[f_id]
            # If "defining_factor" exists in the processor, ok. If not, create it.
            # Find factor_type in the processor
            factor_name = get_processor_id(o.factor.processor) + ":" + factor_type
            factors[factor_name]
            e.append((factor_name, f_id, {"w": o.value.expression, "label": o.value.expression, **edge_factors_relative_to}))

    # Directed Flows between Factors
    for df in objs[FactorsRelationDirectedFlowObservation]:
        sf = get_factor_id(df.source_factor)
        tf = get_factor_id(df.target_factor)
        # Check that both "sf" and "tf" exist in the nodes list (using "factors")
        factors[sf]
        factors[tf]
        weight = df.weight if df.weight else "1"
        e.append((sf, tf, {"w": weight, "label": weight, **edge_factors_flow}))

    # TODO Consider Upscale relations
    # e.append((..., ..., {"w": upscale_weight, "label": upscale_weight, **edge_factors_upscale}))

    # -- Create the graph
    factors_graph = nx.DiGraph()
    factors_graph.add_nodes_from(n)
    factors_graph.add_edges_from(e)

    # nx.write_gml(factors_graph, "/home/rnebot/IntermediateGraph.gml")

    # ---- AUXILIARY GRAPH: FACTOR TYPES AND THEIR INTERRELATIONS ----
    n = []
    e = []
    # --   Nodes: "FactorType"s passing the filter
    for ft in objs[FactorType]:
        n.append((get_factor_type_id(ft), dict(factor_type=ft)))

    # --   Edges
    # Hierarchy and expressions stated in the hierarchy
    ft_in = {}  # Because FactorTypes cannot be both in hierarchy AND expression, marks if it has been specified one was, to raise an error if it is specified also the other way
    for ft in objs[FactorType]:
        ft_id = get_factor_type_id(ft)
        if ft.expression:
            if ft not in ft_in:
                # TODO Create one or more relations, from other FactorTypes (same Hierarchy) to this one
                # TODO The expression can only be a sum of FactorTypes (same Hierarchy)
                ft_in[ft] = "expression"
                # TODO Check that both "ft-id" and "..." exist in the nodes list (keep a temporary set)
                # weight = ...
                # e.append((ft_id, ..., {"w": weight, "label": weight, "origin": ft, "destination": ...}))

        if ft.parent:
            if ft.parent not in ft_in or (ft.parent in ft_in and ft_in[ft.parent] == "hierarchy"):
                # Create an edge from this FactorType
                ft_in[ft.parent] = "hierarchy"
                parent_ft_id = get_factor_type_id(ft.parent)
                # TODO Check that both "ft-id" and "parent_ft_id" exist in the nodes list (keep a temporary set)
                # Add the edge
                e.append((ft_id, parent_ft_id, {"w": "1", "origin": ft, "destination": ft.parent}))
            else:
                raise Exception("The FactorType '"+ft_id+"' has been specified by an expression, it cannot be parent.")
    # Linear transformations
    for f_rel in objs[FactorTypesRelationUnidirectionalLinearTransformObservation]:
        origin = get_factor_type_id(f_rel.origin)
        destination = get_factor_type_id(f_rel.destination)
        e.append((origin, destination, {"w": f_rel.weight, "label": f_rel.weight, "origin": f_rel.origin, "destination": f_rel.destination}))

    # --   Create FACTOR TYPES graph
    factor_types_graph = nx.DiGraph()
    factor_types_graph.add_nodes_from(n)
    factor_types_graph.add_edges_from(e)

    # ---- Obtain weakly connected components of factor_types_graph ----
    factor_types_subgraphs = list(nx.weakly_connected_component_subgraphs(factor_types_graph))

    # ---- EXPAND FACTORS GRAPH with FACTOR TYPES RELATIONS ----
    # The idea is: clone a FactorTypes subgraph if a Factor instances some of its member nodes
    # This cloning process can imply creating NEW Factors

    the_new_node_names_set = set()

    # Obtain weak components of the main graph. Each can be considered separately

    # for sg in nx.weakly_connected_component_subgraphs(factors_graph):  # For each subgraph
    #     print("--------------------------------")
    #     for n in sg.nodes():
    #         print(n)

    sg_list = []  # List of modified (augmented) subgraphs
    for sg in nx.weakly_connected_component_subgraphs(factors_graph):  # For each subgraph
        sg_list.append(sg)
        # Consider each Factor of the subgraph
        unprocessed_factors = set(sg.nodes())
        while unprocessed_factors:  # For each UNPROCESSED Factor
            tmp = unprocessed_factors.pop()  # Get next unprocessed "factor name"
            if tmp not in factors:  # QQ Observations are in the graph and not in "factors". The same with Processors
                continue
            f_ = factors[tmp]  # Obtain Factor from "factor name"
            ft_id = get_factor_type_id(f_)  # Obtain FactorType name from Factor
            # Iterate through FactorTypes and check if the Factor appears
            for sg2 in factor_types_subgraphs:  # Each FactorTypes subgraph
                if ft_id in sg2:  # If the current Factor is in the subgraph
                    if len(sg2.nodes()) > 1:  # If the FactorType subgraph has at least two nodes
                        # CLONE FACTOR TYPES SUBGRAPH
                        # Nodes. Create if not present already
                        n = []
                        e = []
                        for n2, attrs in sg2.nodes().items():  # Each node in the FactorTypes subgraph
                            ft_ = attrs["factor_type"]
                            f_id = get_factor_id(f_.processor, ft_)
                            if f_id not in sg:  # If the FactorType is not
                                # Create Factor, from processor and ft_ -> f_new
                                _, _, f_new = find_or_create_observable(state, name=f_id, source="solver")
                                factors[f_id] = f_new
                                if f_id not in the_node_names_set:
                                    if will_write:
                                        n.append((f_id, non_stated_factor))
                                    else:
                                        d = dict(factor=factor_to_dict(f_new), observations=[])
                                        n.append((f_id, d))
                                if f_id not in the_node_names_set:
                                    the_new_node_names_set.add(f_id)
                                the_node_names_set.add(f_id)
                            else:
                                unprocessed_factors.discard(f_id)
                        # Edges. Create relations between factors
                        for r2, w_ in sg2.edges().items():
                            # Find origin and destination nodes. Copy weight. Adapt weight? If it refers to a FactorType, instance it?
                            origin = get_factor_id(f_.processor, w_["origin"])
                            destination = get_factor_id(f_.processor, w_["destination"])
                            if origin in the_new_node_names_set or destination in the_new_node_names_set:
                                graphics = edge_from_factor_type
                            else:
                                graphics = {}
                            e.append((origin, destination, {"w": w_["w"], "label": w_["w"], **graphics}))
                        sg.add_nodes_from(n)
                        sg.add_edges_from(e)
                        break

    # for sg in sg_list:
    #     print("--------------------------------")
    #     for n in sg.nodes():
    #         print(n)

    # Recompose the original graph
    if sg_list:
        factors_graph = nx.compose_all(sg_list)
    else:
        factors_graph = nx.DiGraph()

    # ----
    # Add "Processor"s just as a way to visualize grouping of factors (they do not influence relations between factors)
    # -
    if include_processors:
        n = []
        e = []
        for p in objs[Processor]:
            p_id = get_processor_id(p)
            if will_write:
                n.append((p_id, a_processor))
            else:
                n.append((p_id, processor_to_dict(p)))
            # Edges between Processors and Factors
            for f in p_factors[p]:
                f_id = get_factor_id(f)
                e.append((p_id, f_id, edge_processor_to_factor))
        factors_graph.add_nodes_from(n)
        factors_graph.add_edges_from(e)

    #
    # for ft in objs[FactorType]:
    #     if ft.parent:
    #         # Check which Factors are instances of this FactorType
    #         if ft in f_types:
    #             for f in f_types[ft]:
    #                 # Check if the processor contains the parent Factor
    #                 processor_factors = p_factors[f.processor]
    #                 if ft.parent not in processor_factors:
    #                     factor_data = (f.processor, ft)
    #                 else:
    #                     factor_data = None
    #                 create_factor = f in qqs  # If there is some Observation
    #                 create_factor = True # Force creation
    #
    #
    #         # Consider the creation of a relation
    #         # Consider also the creation of a new Factor (a new Node for now): if the child has some observation for sure (maybe a child of the child had an observation, so it is the same)
    #         ft_id =
    #     ft_id =

    # Plot graph to file
    # import matplotlib.pyplot as plt
    # ax = plt.subplot(111)
    # ax.set_title('Soslaires Graph', fontsize=10)
    # nx.draw(factors_graph, with_labels=True)
    # plt.savefig("/home/rnebot/Graph.png", format="PNG")

    # GML File
    # nx.write_gml(factors_graph, "/home/rnebot/Graph.gml")

    # VISJS JSON

    # Assign IDs to nodes. Change edges "from" and "to" accordingly
    ids_map = create_dictionary()
    id_count = 0
    for node in factors_graph.nodes(data=True):
        sid = str(id_count)
        node[1]["id"] = sid
        ids_map[node[0]] = sid
        id_count += 1

    vis_nodes = []
    vis_edges = []
    for node in factors_graph.nodes(data=True):
        d = dict(id=node[1]["id"], label=node[0])
        if "shape" in node[1]:
            # circle, ellipse, database, box, diamond, dot, square, triangle, triangleDown, text, star
            d["shape"] = node[1]["shape"]
        else:
            d["shape"] = "box"
        if "color" in node[1]:
            d["color"] = node[1]["color"]
        vis_nodes.append(d)
    for edge in factors_graph.edges(data=True):
        f = ids_map[edge[0]]
        t = ids_map[edge[1]]
        d = {"from": f, "to": t, "arrows": "to"}
        data = edge[2]
        if "w" in data:
            d["label"] = data["w"]
            d["font"] = {"align": "horizontal"}

        vis_edges.append(d)
    visjs = {"nodes": vis_nodes, "edges": vis_edges}
    return visjs

# #########################################################################3

    # GEXF File
    # nx.write_gexf(factors_graph, "/home/rnebot/Graph.gexf")

    # Legend graph
    n = []
    e = []
    n.append(("Factor with Observation", stated_factor_some_observation))
    n.append(("Factor with No Observation", stated_factor_no_observation))
    if include_processors:
        n.append(("Processor", a_processor))
    n.append(("Factor from FactorType", non_stated_factor))
    n.append(("QQ Observation", qq_attached_to_factor))
    n.append(("QQ Intensive Observation", qq_attached_to_factor))

    e.append(("A Factor", "Another Factor", {"label": "Flow between Factors, attaching the weight", **edge_factors_flow}))
    e.append(("Factor #1", "Factor #2", {"label": "Relation from a FactorType", **edge_from_factor_type}))
    if include_processors:
        e.append(("Processor", "A Factor", {"label": "Link from Processor to Factor", **edge_processor_to_factor}))
    e.append(("A Factor", "Same Factor in another processor", {"label": "Upscale a Factor in two processors", **edge_factors_upscale}))
    e.append(("Factor with Observation", "QQ Intensive Observation", {"label": "Observation proportional to extensive value of factor same processor", **edge_factors_relative_to}))
    e.append(("QQ Observation", "A Factor", {"label": "A QQ Observation", **edge_factor_value}))
    factors_graph = nx.DiGraph()
    factors_graph.add_nodes_from(n)
    factors_graph.add_edges_from(e)
    # nx.write_gml(factors_graph, "/home/rnebot/LegendGraph.gml")


if __name__ == '__main__':
    # Deserialize previously recorded Soslaires State (WARNING! Execute unit tests to generated the ".serialized" file)
    with open("/home/rnebot/GoogleDrive/AA_MAGIC/Soslaires.serialized", "r") as file:
        s = file.read()
    state = deserialize_state(s)
    # Create a Query and execute a query
    query = BasicQuery(state)
    construct_flow_graph(state, query, None)


# NUTS files
# Each file is a Code List
# Levels. Each level has a name
# Entry.
#   Each code has a shape (reproject (unify) to geographic coordinates (WKT)).
#   Each code can have a parent.
#   Each code is in a level.
#   An entry can have several codes and descriptions

#
# import shapefile
# dbf = "NUTS_RG_BN_60M_2013"
# # "NUTS_AT_2013",
# lst = ["NUTS_LB_2013",
#        "NUTS_SEPA_LI_2013",
#        "NUTS_RG_60M_2013",
#        "NUTS_JOIN_LI_2013",
#        "NUTS_BN_60M_2013"
#        ]
# base_dir = "/home/rnebot/Downloads/borrame/NUTS_2013_60M_SH/data/"
# for f in lst:
#     sh = shapefile.Reader(base_dir + f)
#     n_shapes = len(sh.shapes())
#     print("-----------------------------------------")
#     print(f + " #" + str(n_shapes))
#     shapes = sh.shapes()
#     fields = sh.fields
#     records = sh.records()
#     print("-----------------------------------------")
#     for i in range(10):
#         s_type = sh.shape(i).shapeType
#         if s_type == 1:
#             bbox = [0, 0, 0, 0]
#         else:
#             bbox = sh.shape(i).bbox
#         n_points = len(sh.shape(i).points)
#         print(str(s_type)+"; # points:" + str(n_points) + "; " + str(['%.3f' % coord for coord in bbox]))



from typing import Dict, Any

from nexinfosys.command_executors import BasicCommand
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import IType
from nexinfosys.common.helper import strcmp
from nexinfosys.models.musiasem_concepts import Indicator, IndicatorCategories, Benchmark


class ScalarIndicatorsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        """
        Create and register Indicator object

        :param fields:
        """
        benchmark_names = fields["benchmarks"]

        benchmarks = []
        if benchmark_names:
            for benchmark_name in benchmark_names.split(","):
                if benchmark_name:
                    benchmark = self._glb_idx.get(Benchmark.partial_key(benchmark_name))
                    if len(benchmark) == 1:
                        benchmark = benchmark[0]
                    elif len(benchmark) == 0:
                        self._add_issue(IType.ERROR,
                                        f"Benchmark {benchmark_name} does not exist (it must be declared previously in a "
                                        "ScalarBenchmark command worksheet")
                        return
                    elif len(benchmark) > 1:
                        self._add_issue(IType.ERROR,
                                        f"Benchmark {benchmark_name} exists {len(benchmark)} times."
                                        " Only one occurrence is allowed.")
                        return
                else:
                    benchmark = None
                if benchmark:
                    benchmarks.append(benchmark)

        indicator = Indicator(fields["indicator_name"],
                              fields["formula"],
                              None,
                              fields.get("processors_selector"),  # Not gathered from worksheet
                              benchmarks,
                              IndicatorCategories.factors_expression if strcmp(fields.get("local"), "Yes")
                              else IndicatorCategories.case_study,
                              fields.get("description"),
                              fields["indicators_group"],
                              fields["unit"],
                              fields["unit_label"],
                              fields["source"],
                              fields["account_na"])
        self._glb_idx.put(indicator.key(), indicator)



import json

from nexinfosys.models.musiasem_concepts import PedigreeMatrix
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects


class PedigreeMatrixCommand(IExecutableCommand):
    """
    Declaration of a pedigree matrix, which can be used

    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        Create the PedigreeMatrix object to which QQ observation may refer

        """
        some_error = False
        issues = []

        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

        # Prepare PedigreeMatrix object instance
        phases = self._content["phases"]
        codes = self._content["codes"]
        name = self._content["name"]

        pm = PedigreeMatrix(name)
        pm.set_phases(codes, phases)

        # Insert the PedigreeMatrix object into the state
        glb_idx.put(pm.key(), pm)
        import jsonpickle
        s = jsonpickle.encode(pm)
        pm2 = jsonpickle.decode(s)

        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



import re
from typing import Dict, Any

from nexinfosys.command_executors import CommandExecutionError, BasicCommand, subrow_issue_message
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.command_generators import IType
from nexinfosys.command_generators.parser_field_parsers import string_to_ast, processor_names
from nexinfosys.common.helper import strcmp
from nexinfosys.models.musiasem_concepts import Factor, RelationClassType, Processor, \
    FactorTypesRelationUnidirectionalLinearTransformObservation, Observer
from nexinfosys.models.musiasem_concepts_helper import create_relation_observations, \
    find_factor_types_transform_relation, find_or_create_observer
from nexinfosys.solving import get_processor_names_to_processors_dictionary


class RelationshipsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))
        self._all_processors = None

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        def process_relation(relation_class):
            source_processor = self._get_processor_from_field("source_processor")
            target_processor = self._get_processor_from_field("target_processor")

            self._check_fields(relation_class, source_processor, target_processor, subrow)

            if relation_class.is_between_processors:
                create_relation_observations(self._glb_idx, source_processor, [(target_processor, relation_class)],
                                             relation_class, None, attributes=attributes)

            elif relation_class.is_between_interfaces:
                try:
                    source_interface = self._get_interface_from_field("source_interface", source_processor) if self._fields.get("source_interface") else self._get_interface_from_field("target_interface", source_processor)
                except CommandExecutionError as e:
                    source_interface = None
                    if not str(e).startswith("The interface"):
                        raise e
                    else:
                        self._add_issue(IType.WARNING, str(e))

                try:
                    target_interface = self._get_interface_from_field("target_interface", target_processor) if self._fields.get("target_interface") else self._get_interface_from_field("source_interface", target_processor)
                except CommandExecutionError as e:
                    target_interface = None
                    if not str(e).startswith("The interface"):
                        raise e
                    else:
                        self._add_issue(IType.WARNING, str(e))

                if not source_interface or not target_interface:
                    return

                if fields["back_interface"]:
                    relation_class = RelationClassType.ff_directed_flow_back

                if relation_class == RelationClassType.ff_directed_flow_back:
                    back_interface = self._get_interface_from_field("back_interface", source_processor)
                    self._check_flow_back_interface_types(source_interface, target_interface, back_interface)
                    attributes.update(dict(back_interface=back_interface))

                if relation_class.is_flow:
                    self._check_flow_orientation(
                        source_processor, target_processor, source_interface, target_interface,
                        is_direct_flow=(relation_class == RelationClassType.ff_directed_flow)
                    )

                if source_interface.taxon != target_interface.taxon:
                    interface_types_transforms = find_factor_types_transform_relation(
                        self._glb_idx, source_interface.taxon, target_interface.taxon, source_processor, target_processor)

                    # ChangeOfTypeScale
                    if self._fields.get("change_type_scale"):
                        o = FactorTypesRelationUnidirectionalLinearTransformObservation.create_and_append(
                            source_interface.taxon, target_interface.taxon, self._fields.get("change_type_scale"),
                            source_interface.processor, target_interface.processor,  # AdHoc source-target Context
                            None, None,  # No unit conversion
                            find_or_create_observer(Observer.no_observer_specified, self._glb_idx))
                        self._glb_idx.put(o.key(), o)
                        if len(interface_types_transforms) > 0:
                            self._add_issue(IType.WARNING,
                                            f"Preexisting matching ScaleChangeMap entry found. Overriding with "
                                            f"{self._fields.get('change_type_scale')}")

                    interface_types_transform = self._get_interface_types_transform(
                        source_interface.taxon, source_processor, target_interface.taxon, target_processor, subrow)
                    attributes.update(dict(scale_change_weight=interface_types_transform.scaled_weight))

                create_relation_observations(self._glb_idx, source_interface,
                                             [(target_interface, relation_class, fields["flow_weight"])],
                                             relation_class, None, attributes=attributes)

        if not self._all_processors:
            self._all_processors = get_processor_names_to_processors_dictionary(self._glb_idx)
        # source_cardinality = fields["source_cardinality"]
        # target_cardinality = fields["target_cardinality"]
        source_processors = self._fields["source_processor"]
        target_processors = self._fields["target_processor"]
        attributes = self._get_attributes_from_field("attributes")

        try:  # Get relation class type
            relation_class = RelationClassType.from_str(fields["relation_type"])
        except NotImplementedError as e:
            raise CommandExecutionError(str(e))

        if ".." in source_processors or ".." in target_processors:
            if ".." in source_processors:
                source_processor_names = obtain_matching_processors(string_to_ast(processor_names, self._fields["source_processor"]), self._all_processors)
            else:
                source_processor_names = [source_processors]
            if ".." in target_processors:
                target_processor_names = obtain_matching_processors(string_to_ast(processor_names, self._fields["target_processor"]), self._all_processors)
            else:
                target_processor_names = [target_processors]
            for s in source_processor_names:
                for t in target_processor_names:
                    self._fields["source_processor"] = s
                    self._fields["target_processor"] = t
                    process_relation(relation_class)
        else:
            process_relation(relation_class)

    def _check_fields(self, relation_class: RelationClassType, source_processor: Processor, target_processor: Processor, subrow=None):
        # Use of column BackInterface is only allowed in some relation types
        back_allowed_classes = [RelationClassType.ff_directed_flow, RelationClassType.ff_reverse_directed_flow,
                                RelationClassType.ff_directed_flow_back]
        if self._fields["back_interface"] and relation_class not in back_allowed_classes:
            raise CommandExecutionError(f"Column 'BackInterface' is only allowed in relations of type: "
                                        f"{back_allowed_classes}"+subrow_issue_message(subrow))

        # Use of column Weight is only allowed in some relation types
        weight_allowed_classes = [RelationClassType.ff_directed_flow, RelationClassType.ff_reverse_directed_flow,
                                  RelationClassType.ff_directed_flow_back, RelationClassType.ff_scale]
        if self._fields["flow_weight"] and relation_class not in weight_allowed_classes:
            raise CommandExecutionError(f"Column 'Weight' is only allowed in relations of type: "
                                        f"{weight_allowed_classes}")

        # Processors should be the same when relation is "Scale Change"
        if relation_class == RelationClassType.ff_scale_change and source_processor.name != target_processor.name:
            raise CommandExecutionError(
                f"Source and target processors should be the same for a relation of type"
                f" 'Scale Change': '{source_processor.name}' != '{target_processor.name}'")

        # TODO: all flows from same "proc:iface" to same "proc" should have the same Weight

    def _check_flow_back_interface_types(self, source: Factor, target: Factor, back: Factor):
        if source.taxon == target.taxon:
            raise CommandExecutionError(f"The type of source and target interfaces should be different. "
                                        f"Source and Target: {source.taxon.name}")

        if back.taxon != target.taxon:
            raise CommandExecutionError(f"The type of target and back interfaces should be the same. Target: "
                                        f"{target.taxon.name}; Back: {back.taxon.name}")

    def _check_flow_orientation(self, source_processor: Processor, target_processor: Processor,
                                source_interface: Factor, target_interface: Factor, is_direct_flow: bool):
        """Check for correct interfaces orientation (input/output) of source and target"""
        allowed_source_orientation = ("Output" if is_direct_flow else "Input")

        # Are the orientations equal?
        if strcmp(source_interface.orientation, target_interface.orientation):
            if strcmp(source_interface.orientation, allowed_source_orientation):
                # Target processor should be parent of source processor
                parent_processor, child_processor = target_processor, source_processor
            else:
                # Source processor should be parent of target processor
                parent_processor, child_processor = source_processor, target_processor

            if child_processor not in parent_processor.children(self._glb_idx):
                raise CommandExecutionError(f"The processor '{child_processor.name}' should be part of the "
                                            f"processor '{parent_processor.name}' when using the same interface "
                                            f"orientation '{source_interface.orientation}'.")

        else:  # Orientations are different
            if not strcmp(source_interface.orientation, allowed_source_orientation):
                raise CommandExecutionError(f"The source interface '{source_interface.full_name}' has the wrong "
                                            f"orientation '{source_interface.orientation}'.")

            if strcmp(target_interface.orientation, allowed_source_orientation):
                raise CommandExecutionError(f"The target interface '{target_interface.full_name}' has the wrong "
                                            f"orientation '{target_interface.orientation}'.")

    # def parse_and_unfold_line(self, item, hh, datasets, parameters, all_processors):
    #     # Consider multiplicity because of:
    #     # - A dataset (only one). First a list of dataset concepts used in the line is obtained.
    #     #   Then the unique tuples formed by them are obtained.
    #     # - Processor name.
    #     #   - A set of processors (wildcard or filter by attributes)
    #     #   - A set of interfaces (according to another filter?)
    #     # - Multiple types of relation
    #     # - Both (first each dataset record applied -expanded-, then the name evaluation is applied)
    #     # - UNRESOLVED: expressions are resolved partially. Parts where parameters
    #     # expressions depending on parameters. Only the part of the expression depending on varying things
    #     # - The processor name could be a concatenation of multiple literals
    #     #
    #     # Look for multiple items in r_source_processor_name, source_interface_name,
    #     #                            r_target_processor_name, target_interface_name
    #     if item["_complex"]:
    #         asts = parse_line(item, self._command_fields)
    #         if item["_expandable"]:
    #             # It is an expandable line
    #             # Look for fields which are specified to be variable in order to originate the expansion
    #             res = classify_variables(asts, datasets, hh, parameters)
    #             ds_list = res["datasets"]
    #             ds_concepts = res["ds_concepts"]
    #             h_list = res["hierarchies"]
    #             if len(ds_list) >= 1 and len(h_list) >= 1:
    #                 self._add_issue(IType.ERROR, "Dataset(s): "+", ".join([d.name for d in ds_list])+", and hierarchy(ies): "+", ".join([h.name for h in h_list])+", have been specified. Either a single dataset or a single hiearchy is supported.")
    #                 return
    #             elif len(ds_list) > 1:
    #                 self._add_issue(IType.ERROR, "More than one dataset has been specified: "+", ".join([d.name for d in ds_list])+", just one dataset is supported.")
    #                 return
    #             elif len(h_list) > 1:
    #                 self._add_issue(IType.ERROR, "More than one hierarchy has been specified: " + ", ".join([h.name for h in h_list])+", just one hierarchy is supported.")
    #                 return
    #             const_dict = obtain_dictionary_with_literal_fields(item, asts)
    #             if len(ds_list) == 1:
    #                 # If a measure is requested and not all dimensions are used, aggregate or
    #                 # issue an error (because it is not possible to reduce without aggregation).
    #                 # If only dimensions are used, then obtain all the unique tuples
    #                 ds = ds_list[0]
    #                 measure_requested = False
    #                 all_dimensions = set([c.code for c in ds.dimensions if not c.is_measure])
    #                 for con in ds_concepts:
    #                     for c in ds.dimensions:
    #                         if strcmp(c.code, con):
    #                             if c.is_measure:
    #                                 measure_requested = True
    #                             else:  # Dimension
    #                                 all_dimensions.remove(c.code)
    #                 only_dimensions_requested = len(all_dimensions) == 0
    #
    #                 if measure_requested and not only_dimensions_requested:
    #                     self._add_issue(IType.ERROR, "It is not possible to use a measure if not all dimensions are used (cannot assume implicit aggregation)")
    #                     return
    #                 elif not measure_requested and not only_dimensions_requested:
    #                     # TODO Reduce the dataset to the unique tuples (consider the current case -sensitive or not-sensitive-)
    #                     data = None
    #                 else:  # Take the dataset as-is
    #                     data = ds.data
    #
    #                 for row in data.iterrows():
    #                     item2 = const_dict.copy()
    #
    #                     d = {}
    #                     for c in ds_concepts:
    #                         d["{" + ds.code + "." + c + "}"] = row[c]
    #                     # Expand in all fields
    #                     for f in self._command_fields:
    #                         if f not in const_dict:
    #                             # Replace all
    #                             string = item[f]
    #                             # TODO Could iterate through the variables in the field (not IN ALL FIELDS of the row)
    #                             for item in sorted(d.keys(), key=len, reverse=True):
    #                                 string = re.sub(item, d[item], string)
    #                             item2[f] = string
    #                     # Now, look for wildcards where it is allowed
    #                     r_source_processor_name = string_to_ast(processor_names, item2.get("source_processor"))
    #                     r_target_processor_name = string_to_ast(processor_names, item2.get("target_processor"))
    #                     if ".." in r_source_processor_name or ".." in r_target_processor_name:
    #                         if ".." in r_source_processor_name:
    #                             source_processor_names = obtain_matching_processors(r_source_processor_name, all_processors)
    #                         else:
    #                             source_processor_names = [r_source_processor_name]
    #                         if ".." in r_target_processor_name:
    #                             target_processor_names = obtain_matching_processors(r_target_processor_name, all_processors)
    #                         else:
    #                             target_processor_names = [r_target_processor_name]
    #                         for s in source_processor_names:
    #                             for t in target_processor_names:
    #                                 item3 = item2.copy()
    #                                 item3["source_processor"] = s
    #                                 item3["target_processor"] = t
    #                                 print("Multiple by dataset and wildcard: " + str(item3))
    #                                 yield item3
    #                     else:
    #                         print("Multiple by dataset: " + str(item3))
    #                         yield item2
    #             elif len(h_list) == 1:
    #                 pass
    #             else:  # No dataset, no hierarchy of categories, but still complex, because of wildcards
    #                 wildcard_in_source = ".." in item.get("source_processor", "")
    #                 wildcard_in_target = ".." in item.get("target_processor", "")
    #                 if wildcard_in_source or wildcard_in_target:
    #                     r_source_processor_name = string_to_ast(processor_names, item.get("source_processor"))
    #                     r_target_processor_name = string_to_ast(processor_names, item.get("target_processor"))
    #                     if wildcard_in_source:
    #                         source_processor_names = obtain_matching_processors(r_source_processor_name, all_processors)
    #                     else:
    #                         source_processor_names = [item["source_processor"]]
    #                     if wildcard_in_target:
    #                         target_processor_names = obtain_matching_processors(r_target_processor_name, all_processors)
    #                     else:
    #                         target_processor_names = [item["target_processor"]]
    #                     for s in source_processor_names:
    #                         for t in target_processor_names:
    #                             item3 = const_dict.copy()
    #                             item3["source_processor"] = s
    #                             item3["target_processor"] = t
    #                             print("Multiple by wildcard: "+str(item3))
    #                             yield item3
    #                 else:
    #                     # yield item
    #                     raise Exception("If 'complex' is signaled, it should not pass by this line")
    #     else:
    #         # print("Single: "+str(item))
    #         yield item

    # def execute(self, state: "State") -> IssuesOutputPairType:
    #     self._init_execution_state(state)
    #     self._glb_idx, _, hh, datasets, _ = get_case_study_registry_objects(state)
    #
    #     # Obtain the names of all parameters
    #     parameters = [p.name for p in self._glb_idx.get(Parameter.partial_key())]
    #
    #     # Obtain the names of all processors
    #     all_processors = get_processor_names_to_processors_dictionary(self._glb_idx)
    #
    #     for row in self._content["items"]:
    #         for sub_row in self.parse_and_unfold_line(row, hh, datasets, parameters, all_processors):
    #             try:
    #                 self._init_and_process_row(sub_row)
    #             except CommandExecutionError as e:
    #                 self._add_issue(IType.ERROR, str(e))
    #
    #     return self._issues, None


def obtain_matching_processors(parsed_processor_name, all_processors):
    """

    :param parsed_processor_name: The AST of parsing processor names (rule "processor_names")
    :param all_processors: either a list with all processor names or a dict with full processor name to Processor
    :return: either the set of processor names matching the filter or the set of Processor whose names match the filter
    """
    # Prepare "processor_name"
    s = r""
    first = True
    for p in parsed_processor_name["parts"]:
        if p[0] == "separator":
            if p[1] == "..":
                if first:
                    s += r".*"
                    if len(parsed_processor_name["parts"]) > 1:
                        s += r"\."
                else:
                    s += r"\..*"
            else:
                s += r"\."
        else:
            s += p[1]
        first = False
    reg = re.compile(s)
    res = set()
    add_processor = isinstance(all_processors, dict)
    for p in all_processors:
        if reg.match(p):
            if add_processor:
                res.add(all_processors[p])
            else:
                res.add(p)

    return res



import json

from nexinfosys.models.musiasem_concepts_helper import build_hierarchy
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects


class HierarchyCommand(IExecutableCommand):
    """
    Serves to specify a hierarchy (composition or taxonomy) of Observables
    Observables can be Processors, Factors or external categories
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        Create a Hierarchy. The exact form of this hierarchy is different depending on the concept:
        * FactorTypes and Categories use Hierarchies, which are intrinsic.
            The hierarchy name is passed to the containing Hierarchy object
        * Processors use Part-Of Relations. In this case, the hierarchy name is lost
        Names of Processor and FactorTypes are built both in hierarchical and simple form
        The hierarchical is all the ancestors from root down to the current node, separated by "."
        The simple name is just the current node. If there is already another concept with that name, the simple name
        is not stored (STORE BOTH CONCEPTS by the same name, and design some tie breaking mechanism??)
        """
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        # Extract base information
        name = self._content["name"]
        type_name = self._content["type"]
        first_level = self._content["h"]

        # Build the hierarchy
        build_hierarchy(name, type_name, registry=glb_idx, h=first_level)

        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



import json
from io import StringIO
import pandas as pd

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.common.helper import strcmp, prepare_dataframe_after_external_read, create_dictionary, any_error_issue, \
    translate_case
from nexinfosys.models import CodeImmutable
from nexinfosys.models.statistical_datasets import CodeList


class DatasetDataCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        issues = []

        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        name = self._content["command_name"]

        # List of available dataset names. The newly defined datasets must not be in this list
        ds_names = [ds.code for ds in datasets.values()]

        # List of datasets with local worksheet name
        external_dataset_names = []
        for ds in datasets.values():
            if ds.attributes["_location"].lower().startswith("data://#"):
                worksheet = ds.attributes["_location"][len("data://#"):]
                if not worksheet.lower().startswith("datasetdata "):
                    worksheet = "DatasetData " + worksheet

                if strcmp(worksheet, name):
                    external_dataset_names.append(ds.code)

        # Process parsed information
        for r, line in enumerate(self._content["items"]):
            # A dataset
            dataset_names = line["name"]
            if dataset_names == "":
                if external_dataset_names:
                    dataset_names = external_dataset_names
                else:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The column name 'DatasetName' was not defined for command 'DatasetData' and there is no 'location' in a DatasetDef command pointing to it",
                                        location=IssueLocation(sheet_name=name, row=1, column=None)))
            else:
                dataset_names = [dataset_names]

            # Find it in the already available datasets. MUST EXIST
            for n in ds_names:
                for dataset_name in dataset_names:
                    if strcmp(dataset_name, n):
                        df = pd.read_json(StringIO(line["values"]), orient="split")
                        # Check columns
                        ds = datasets[n]
                        iss = prepare_dataframe_after_external_read(ds, df, name)
                        issues.extend(iss)
                        # Everything ok? Store the dataframe!
                        if not any_error_issue(iss):
                            r = ds.attributes["_dataset_first_row"]
                            # Loop over "ds" concepts.
                            # - "dimension" concepts of type "string" generate a CodeHierarchy
                            # - Check that the DataFrame contains ALL declared concepts. If not, generate issue
                            # dims = translate_case([d.code for d in ds.dimensions], df.columns)
                            cid = create_dictionary(data={col: col for col in df.columns})
                            col_names = list(df.columns)
                            for c in ds.dimensions:
                                if c.code in cid:
                                    col_names[df.columns.get_loc(cid[c.code])] = c.code  # Rename column
                                    dsd_concept_data_type = c.attributes["_datatype"]
                                    if dsd_concept_data_type.lower() == "string" and not c.is_measure:  # Freely defined dimension
                                        cl = df[cid[c.code]].unique().tolist()
                                        c.code_list = CodeList.construct(
                                            c.code, c.code, [""],
                                            codes=[CodeImmutable(c, c, "", []) for c in cl]
                                        )
                                else:
                                    issues.append(Issue(itype=IType.ERROR,
                                                        description=f"Concept '{c.code}' not defined for '{ds.code}'",
                                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                            df.columns = col_names
                            ds.data = df
                        dataset_names.remove(dataset_name)
                        break

            if dataset_names:
                issues.append(
                    Issue(itype=IType.ERROR,
                          description=f"Metadata for the datasets: {','.join(dataset_names)}, must be defined previously",
                          location=IssueLocation(sheet_name=name, row=-1, column=-1)))

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues



from nexinfosys.command_generators.parser_spreadsheet import commands_generator_from_ooxml_file
from nexinfosys.command_generators.parser_json import commands_generator_from_native_json

"""
- Parser
  - Receives a stream which contains one or more command_executors, creates one or more instances of command_executors

- CommandsExecutor
  - Receives a Workspace and one or more command instances to be executed

- Command.
  - Can be created directly, by a Parser (different Parsers possible), or by deserialization
  - Serializable/deserializable in: JSON or DataFrame format. Execute. Acting always over a
  - Could have a method to estimate the execution time

- Workspace
  - Can be serialized/deserialized
  - Allows reading, writing, deleting variables
  - Can integrate the result of a Command (if the Command does not have this capability)

"""


def commands_container_parser_factory(generator_type, file_type, file, state, sublist=None, stack=None, ignore_imports=False):
    """
    Returns a generator appropriate to parse "file" and generate command_executors

    :param generator_type:
    :param file_type:
    :param file:
    :param state: State used to validate existence of some variables at parse time
    :param sublist: Sublist of worksheets to use. If empty, all worksheets
    :param stack: Stack of workbook "hashes" to guide the recursive parsing process and to avoid circular imports
    :return:
    """
    def hash_file(f):
        import hashlib
        m = hashlib.md5()
        if isinstance(f, str):
            m.update(f.encode("utf-8"))
        else:
            m.update(f)
        return m.digest()

    if not stack:
        stack = []

    # TODO Prepare the input stream. It can be a String, a URL, a file, a Stream
    # TODO Define a routine to read it into memory
    s = file

    h = hash_file(s)
    for i in range(len(stack)):
        if stack[i] == h:
            # TODO Another possibility is to allow the circular reference, simply ignoring it (in this point,
            #  properly warn (use yield), then return to end this level)
            raise Exception("Circular reference detected importing file")
    # Push if OK, then process the file
    stack.append(h)

    if generator_type.lower() in ["rscript", "r-script", "r"]:
        # TODO The R script was prepared to be run from outside NIS, using R NIS client
        # TODO Running the script from the inside should be managed slightly different:
        # TODO - Recognize that it is an internally launched script
        # TODO   - The R script will open an interactive session: do not open a new InteractiveSession and
        # TODO   - find a way to reenter the launching Int.Sess.
        # TODO   - ignore open/close session commands creating or saving case studies
        # TODO   - execute commands modifying in memory state, ignore others
        #
        # TODO Take the R script and launch it as a separate process.
        # TODO
        pass
    elif generator_type.lower() in ["python", "python-script"]:
        # TODO Exact same considerations as for R scripts
        pass
    elif generator_type.lower() in ["spreadsheet", "excel", "workbook"]:
        # A sequence of commands, providing the whole case study
        if isinstance(s, bytes):
            pass
        elif isinstance(s, str):
            pass  # TODO It may be a file name

        yield from commands_generator_from_ooxml_file(s, state, sublist, stack, ignore_imports)
    elif generator_type.lower() in ["json", "native", "primitive"]:  # "primitive" is Deprecated
        # A list of commands. Each command is a dictionary: the command type, a label and the content
        # The type is for the factory to determine the class to instantiate, while label and content
        # are passed to the command constructor to elaborate the command
        if isinstance(s, bytes):
            s = s.decode("utf-8")
        yield from commands_generator_from_native_json(s, state, sublist, stack)

    # Pop the file hash from the stack
    stack.pop()



import regex as re
from openpyxl.worksheet.worksheet import Worksheet
from pint.errors import UndefinedUnitError

from nexinfosys.command_generators.parser_field_parsers import simple_ident
from nexinfosys.common.helper import create_dictionary, case_sensitive
from nexinfosys import ureg, AreaTupleType, IssuesLabelContentTripleType
from nexinfosys.command_generators import parser_field_parsers
from nexinfosys.models.musiasem_concepts import allowed_ff_types

"""
User Interface. There will be a special page for command_executors reference.
 * Popup. Search. Copy, paste into the Worksheet
 * Submit command
User Interface. how navigation works. Is it compatible with editors like "Bootstrap Studio", "PineWorks", "Pingendo"?
User Interface. there will be a special page to browse Datasets and their input parameters

All command_executors store the content as JSON loads or similar. Use that for the "serialize" command.
 
* Transform columns into processors with observations
* Need JSONLink to create pointers:
  * Each object needs an ID
  * Pointers point to IDs
  * Dictionary from ID to Location of JSON

* JSONSchema to validate that the structure generated internally is correct
* A graph of Observables
* Observables having Observations
* CALCULATED OBSERVATIONS. Check existence. If not, look for a formula or set of rules matching the requested Observation
  * Then, there must be a repository of rules
* A structure where it is easy to ADD new observations

OUTPUT COMMANDS
* Workbook: create workbook. Pivot table, processors.
* Graph format:  
* Online: browse pivot table, processor relations
"""


def parse_data_input_command(sh: Worksheet, area: AreaTupleType, processors_type: str, state=None) -> IssuesLabelContentTripleType:
    """
    Scans the "area" of input worksheet "sh" where it is assumed a "data input" command
    is present.

    It obtains a list of observations, a list of processors, a list of observables, a list of tags
    All those are represented in JSON format

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :param processors_type: Name for the type of processors. Also label of the command
    :param state: Transient state useful for checking existence of variables
    :return: DataInputCommand, list of issues (issue_type, message)
    """
    some_error = False
    issues = []
    # Define a set of observations (qualified quantities) of observables
    # This set can be replicated. So, ?how to refer to each replica?
    # Regular expression, internal name, Mandatory (True|False)
    known_columns = [(r"Name|Processor[_ ]name", "processor", False),
                     (r"Level", "level", False),
                     (r"Parent", "parent", False),
                     (r"FF[_ ]type", "ff_type", True),
                     (r"Var|Variable", "factor", True),
                     (r"Value|NUSAP\.N", "value", False),  # If value is not specified, then just declare the Factor
                     (r"Unit|NUSAP\.U", "unit", True),  # If blank, a dimensionless amount is assumed
                     (r"Relative[_ ]to", "relative_to", False),
                     (r"Uncertainty|Spread|NUSAP\.S", "uncertainty", False),
                     (r"Assessment|NUSAP\.A", "assessment", False),
                     (r"Pedigree[_ ]matrix|NUSAP\.PM", "pedigree_matrix", False),
                     (r"Pedigree|NUSAP\.P", "pedigree", False),
                     (r"Time|Date", "time", False),
                     (r"Geo|Geolocation", "geolocation", False),
                     (r"Source", "source", False),
                     (r"Comment|Comments", "comments", False)
                     ]

    label = "Processors " + processors_type

    # First, examine columns, to know which fields are being specified
    # Special cases:
    #   Open columns: the field is specified in the cell togheter with the value. Like "attr1=whatever", instead of a header "attr1" and in a row below, a value "whatever"
    #   Complex values: the value has syntactic rules. Like expressions for both quantities AND qualities (like NUSAP)
    #   References: the field refers to additional information in another worksheet. Unique names or ref holder (worksheet name) plus ref inside the worksheet, would be allowed. Also ref type can disambiguate
    mandatory = {t[1]: t[2] for t in known_columns}
    cre = {}  # Column Regular Expression dictionary (K: regular expression; V: RegularExpression object)
    if not case_sensitive:
        flags = re.IGNORECASE
    else:
        flags = 0
    for kc in known_columns:
        cre[kc[0]] = re.compile(kc[0], flags=flags)
    col_names = {}
    standard_cols = {}  # Internal (standardized) column name to column index in the worksheet (freedom in the order of columns)
    attribute_cols = create_dictionary()  # Not recognized columns are considered freely named categories, attributes or tags
    attributes = []  # List of attributes or tags (keys of the previous dictionary)
    col_allows_dataset = create_dictionary()  # If the column allows the reference to a dataset dimension
    for c in range(area[2], area[3]):
        col_name = sh.cell(row=area[0], column=c).value
        if not col_name:
            continue

        col_name = col_name.replace("\n", " ")
        col_names[c] = col_name

        # Match
        found = False
        for kc in known_columns:
            res = cre[kc[0]].search(col_name)
            if res:
                if kc[1] in standard_cols:
                    issues.append((2, "Cannot repeat column name '" + col_name +"' (" + kc[0] +") in data input command '" + processors_type + "'"))
                else:
                    standard_cols[kc[1]] = c
                    col_names[c] = kc[1]  # Override column name with pseudo column name for standard columns
                    if col_names[c].lower() in ["factor", "value", "time", "geolocation"]:
                        col_allows_dataset[col_names[c]] = True
                    else:
                        col_allows_dataset[col_names[c]] = False
                    found = True
                break
        if not found:
            if col_name not in attribute_cols:
                # TODO Check valid col_names. It must be a valid Variable Name
                attribute_cols[col_name] = c
                attributes.append(col_name)
                col_allows_dataset[col_name] = True
            else:
                issues.append((2, "Cannot repeat column name '" + col_name + "' in data input command '" + processors_type + "'"))

    del cre

    # Check if there are mandatory columns missing

    # TODO There could be combinations of columns which change the character of mandatory of some columns
    # TODO For instance, if we are only specifying structure, Value would not be needed
    print("BORRAME - "+str(known_columns))
    print("BORRAME 2 - "+str(standard_cols))
    for kc in known_columns:
        # "kc[2]" is the flag indicating if the column is mandatory or not
        # col_map contains standard column names present in the worksheet
        if kc[2] and kc[1] not in standard_cols:
            some_error = True
            issues.append((3, "Column name '" + kc[0] +"' must be specified in data input command '" + processors_type + "'"))

    # If there are errors, do not continue
    if some_error:
        return issues, label, None

    processor_attribute_exclusions = create_dictionary()
    processor_attribute_exclusions["scale"] = None  # Exclude these attributes when characterizing the processor
    processor_attributes = [t for t in attributes if t not in processor_attribute_exclusions]

    # SCAN rows
    lst_observations = []  # List of ALL observations. -- Main outcome of the parse operation --

    set_pedigree_matrices = create_dictionary()  # List of pedigree templates
    set_processors = create_dictionary()  # List of processor names
    set_factors = create_dictionary()  # List of factors
    set_taxa = create_dictionary()  # Dictionary of taxa with their lists of values. Useful to return CODE LISTS
    set_referenced_datasets = create_dictionary()  # Dictionary of datasets to be embedded into the result (it is a job of the execution part)
    processors_taxa = create_dictionary()  # Correspondence "processor" -> taxa (to avoid changes in this correspondence)

    dataset_column_rule = parser_field_parsers.dataset_with_column
    values = [None]*area[3]
    # LOOP OVER EACH ROW
    for r in range(area[0] + 1, area[1]):  # Scan rows (observations)
        # Each row can specify: the processor, the factor, the quantity and qualities about the factor in the processor
        #                       It can also specify a "flow+containment hierarchy" relation

        row = {}  # Store parsed values of the row

        taxa = create_dictionary()  # Store attributes or taxa of the row

        referenced_dataset = None  # Once defined in a row, it cannot change!!
        # Scan the row first, looking for the dataset. The specification is allowed in certain columns:
        # attribute_cols and some standard_cols
        already_processed = create_dictionary()
        for c in range(area[2], area[3]):
            if c in col_names:
                value = sh.cell(row=r, column=c).value
                if isinstance(value, str) and value.startswith("#"):
                    col_name = col_names[c]
                    if col_allows_dataset[col_name]:
                        if not referenced_dataset:
                            try:
                                ast = parser_field_parsers.string_to_ast(dataset_column_rule, value[1:])
                                if len(ast["parts"]) == 2:
                                    referenced_dataset = ast["parts"][0]
                                    # Remove the dataset variable. It will be stored in "_referenced_dataset"
                                    value = "#" + ast["parts"][1]
                                else:
                                    some_error = True
                                    issues.append((3, "The first dataset reference of the row must contain the "
                                                      "dataset variable name and the dimension name, row " + str(r)))

                                # Mark as processed
                                already_processed[col_name] = None
                            except:
                                some_error = True
                                issues.append((3, "Column '" + col_name + "' has an invalid dataset reference '" + value + "', in row " + str(r)))
                        else:
                            try:
                                ast = parser_field_parsers.string_to_ast(simple_ident, value[1:])
                                # Mark as processed
                                already_processed[col_name] = None
                            except:
                                some_error = True
                                issues.append((3, "Column '" + col_name + "' has an invalid dataset reference '" + value + "', in row " + str(r)))
                        if col_name in standard_cols:
                            row[col_name] = value
                        else:
                            taxa[col_name] = value

                values[c] = value

        # TODO If the flow type is decomposed, compose it first
        for c in standard_cols:
            if c in already_processed:
                continue

            value = values[standard_cols[c]]

            # != "" or not
            if value is None or (value is not None and value == ""):
                if c == "unit":
                    value = "-"
                if not value:
                    if mandatory[c]:
                        some_error = True
                        issues.append((3, "Column '" + c + "' is mandatory, row " + str(r)))
                    continue  # Skip the rest of the iteration!

            # Parse the value
            if c in ["processor", "factor"]:
                # Check that it is a variable name, and allow hierarchical names
                parser_field_parsers.string_to_ast(parser_field_parsers.simple_h_name, value)
            elif c == "pedigree_matrix":
                parser_field_parsers.string_to_ast(parser_field_parsers.simple_ident, value)
            elif c == "relative_to":
                # Two elements, the first a hierarchical name, the second a unit name
                s = value.split(" ")
                if len(s) != 2:
                    some_error = True
                    issues.append((3, "The Relative To value has to have two parts, factor name and unit, separated by a whitespace (specified '"+value+"'), in row "+str(r)))
                else:
                    try:
                        parser_field_parsers.string_to_ast(parser_field_parsers.simple_h_name, s[0])
                    except:
                        some_error = True
                        issues.append((3, "The name specified for the relative to factor '"+s[0]+"' is not valid, in row "+str(r)))

                    # It must be a recognized unit. Check with Pint
                    try:
                        ureg(s[1])
                        ureg.parse_unit_name(s[1], case_sensitive)
                    except UndefinedUnitError:
                        some_error = True
                        issues.append((3, "The unit name '"+s[1]+"' is not registered in the units processing package, in row "+str(r)))
            elif c == "level":
                # A valid level name
                try:
                    parser_field_parsers.string_to_ast(parser_field_parsers.level_name, value)
                except:
                    some_error = True
                    issues.append((3, "The level '" + value + "' syntax is not valid, in row " + str(r)))

            elif c == "parent":
                # Check that value is a valid parent name. It can be either a list of tags OR
                # a processor name, something defining a single processor
                try:
                    parser_field_parsers.string_to_ast(parser_field_parsers.simple_h_name, value)
                except:
                    try:
                        parser_field_parsers.string_to_ast(parser_field_parsers.named_parameters_list, value)
                    except:
                        some_error = True
                        issues.append((3, "Could not parse '"+value+"' as 'parent' in row " + str(r)))
            elif c == "ff_type":
                # The type of flow/fund must be one of a set of possible values. DEFINE THE LIST
                if value.lower() not in allowed_ff_types:
                    some_error = True
                    issues.append((3, "ff_type must be one of :"+', '.join(allowed_ff_types)+", in row " + str(r)))
            elif c == "value":
                if not isinstance(value, str):
                    value = str(value)
                # Expression allowed. Check syntax only. It can refer to parameters.
                ast = parser_field_parsers.string_to_ast(parser_field_parsers.expression, value)
                # TODO Check existence of used variables
                # TODO basic_elements_parser.ast_evaluator(ast, state, None, issues, "static")
            elif c == "unit":
                # It must be a recognized unit. Check with Pint
                try:
                    value = value.replace("€", "Euro").replace("$", "Dollar")
                    if value == "-":
                        value = ""  # Dimensionless
                    ureg(value)
                    ureg.parse_unit_name(value, case_sensitive)
                except:
                    some_error = True
                    issues.append((3, "The unit name '"+value+"' is not registered in the units processing package, in row "+str(r)))
            elif c == "uncertainty":
                # TODO It must be a valid uncertainty specifier
                pass
            elif c == "assessment":
                # See page 135 of Funtowicz S., Ravetz J., "Uncertainty and Quality in Science for Policy"
                # "c" is "cognitive" assessment, "p" is pragmatic assessment.
                allowed = ["nil", "low", "medium", "high", "total",
                           "nil_c", "low_c", "medium_c", "high_c", "total_c",
                           "nil_p", "low_p", "medium_p", "high_p", "total_p"]
                if value and value.lower() not in allowed:
                    issues.append((3, "Assessment must be empty or one of: "+", ".join(allowed)))
            elif c == "pedigree":
                # A valid pedigree specification is just an integer
                try:
                    int(value)
                except:
                    issues.append((3, "The pedigree specification '"+value+"' must be an integer"))
            elif c == "time":
                # A valid time specification. Possibilities: Year, Month-Year / Year-Month, Time span (two dates)
                if not isinstance(value, str):
                    value = str(value)
                ast = parser_field_parsers.string_to_ast(parser_field_parsers.time_expression, value)
            elif c == "geolocation":
                # A reference to a geolocation
                try:
                    parser_field_parsers.string_to_ast(parser_field_parsers.reference, value)
                except:
                    some_error = True
                    issues.append((3, "The geolocation must be a reference"))
            elif c == "source":
                # Who or what provided the information. It can be formal or informal. Formal can be references (but evaluated later)
                pass
            elif c == "comments":
                # Free text
                pass

            # Store the parsed value
            row[c] = value

        for c in attribute_cols:
            if c in already_processed:
                continue

            value = values[attribute_cols[c]]

            # != "" or not
            if not value:
                taxa[c] = None
                continue  # Skip the rest of the iteration!

            # TODO Check value. Valid identifier, no whitespace
            # Validate "value", it has to be a simple ID
            try:
                if not isinstance(value, str):
                    value = str(value)
                parser_field_parsers.simple_ident.parseString(value, parseAll=True)
            except:
                value = None
                some_error = True
                issues.append((3,
                               "The value in column '" + c + "' has to be a simple identifier: start with letter, then letters, numbers and '_', no whitespace, in row " + str(
                                   r)))

            taxa[c] = value

            # Disable the registration of taxa. If a Dataset reference is used, there is no way to register
            # taxa at parse time (the dataset is still not obtained). Leave it for the execution
            if c not in set_taxa:
                set_taxa[c] = create_dictionary()
            if value is not None:
                set_taxa[c][value] = None

        # Now that individual columns have been parsed, do other things

        if referenced_dataset:
            row["_referenced_dataset"] = referenced_dataset

        # If "processor" not specified, concatenate taxa columns in order to generate an automatic name
        # (excluding the processor type)
        p_taxa = taxa.copy()
        for k in processor_attribute_exclusions:
            if k in p_taxa: del p_taxa[k]

        if "processor" not in row:
            row["processor"] = "_".join([str(taxa[t]) for t in processor_attributes])  # TODO Which order? (the current is "order of appearance"; maybe "alphabetical order" would be better option)
        # Add as "taxa" the processor type (which is an optional input parameter to this function)
        if processors_type:
            taxa["_processors_type"] = processors_type
        # Store taxa (attributes and taxa)
        row["taxa"] = taxa
        # Store taxa if the processor still does not have it
        if row["processor"] not in processors_taxa:
            processors_taxa[row["processor"]] = p_taxa  # "::".join([taxa[t] for t in lst_taxa_cols])
        else:
            # Taxa should be the same for each "processor". Error if different
            t = processors_taxa[row["processor"]]
            if t != p_taxa:
                issues.append((3, "The processor '"+row["processor"]+"' has different taxa assigned, in row "+str(r)))

        # Register new processor names, pedigree templates, and variable names
        if "processor" in row:
            set_processors[row["processor"]] = None
        if "pedigree_matrix" in row:
            set_pedigree_matrices[row["pedigree_matrix"]] = None
        if "factor" in row:
            set_factors[row["factor"]] = None
        if referenced_dataset:
            set_referenced_datasets[referenced_dataset] = None

        lst_observations.append(row)

    content = {"factor_observations": lst_observations,
               "processor_attributes": processor_attributes,
               "processors": [k for k in set_processors],
               "pedigree_matrices": [k for k in set_pedigree_matrices],
               "factors": [k for k in set_factors],
               "referenced_datasets": [ds for ds in set_referenced_datasets],
               "code_lists": {k: [k2 for k2 in set_taxa[k]] for k in set_taxa}
               }
    return issues, label, content
    # if not some_error:
    #     cmd = DataInputCommand(label)
    #     cmd.json_deserialize(content)
    # else:
    #     cmd = None
    # return cmd, issues



import json

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.common.helper import strcmp
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.models.musiasem_concepts import HierarchySource, HierarchyGroup, Hierarchy, HierarchyLevel, HierarchyNode, \
    Taxon


class HierarchyCategoriesCommand(IExecutableCommand):
    """
    Serves to specify a hierarchy (composition or taxonomy) of Observables
    Observables can be Processors, Factors or external categories
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        Create a Hierarchy of Taxon. The exact form of this hierarchy is different depending on the concept:
        * FactorTypes and Categories use Hierarchies, which are intrinsic.
            The hierarchy name is passed to the containing Hierarchy object
        * Processors use Part-Of Relations. In this case, the hierarchy name is lost
        Names of Processor and FactorTypes are built both in hierarchical and simple form
        The hierarchical is all the ancestors from root down to the current node, separated by "."
        The simple name is just the current node. If there is already another concept with that name, the simple name
        is not stored (STORE BOTH CONCEPTS by the same name, and design some tie breaking mechanism??)
        """
        issues = []
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        name = self._content["command_name"]

        # Process parsed information
        for item in self._content["items"]:
            r = item["_row"]
            # HierarchySource (Optional)
            hsource = item.get("source", None)  # Code of entity defining the Hierarchy
            if hsource:
                tmp = hsource
                hsource = glb_idx.get(HierarchySource.partial_key(name=hsource))
                if len(hsource) == 0:
                    hsource = HierarchySource(name=tmp)
                    glb_idx.put(hsource.key(), hsource)
                else:
                    hsource = hsource[0]

            hname = item.get("hierarchy_name", None)
            if not hname:
                issues.append(Issue(itype=IType.ERROR,
                                    description="The name of the Hierarchy has not been defined. Skipped.",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                continue

            # HierarchyGroup (equivalent to Hierarchy of Code Lists, HCL)
            hg = item.get("hierarchy_group", None)
            if hg:
                is_code_list = False  # Hierarchy group
            else:
                is_code_list = True  # Hierarchy group for the Code List, with the same name
                hg = hname

            # Check if the HierarchyGroup is previously defined. YES, use it; NO, create new HierarchyGroup
            tmp = hg
            hg = glb_idx.get(HierarchyGroup.partial_key(name=hg))
            if len(hg) == 0:
                hg = HierarchyGroup(name=tmp, source=hsource)
                glb_idx.put(hg.key(), hg)
            else:
                hg = hg[0]

            # Check if the Hierarchy is defined. YES, get it; NO, create it
            tmp = hname
            h = glb_idx.get(Hierarchy.partial_key(name=hname))
            if len(h) == 0:
                h = Hierarchy(name=tmp)
                glb_idx.put(h.key(), h)
                glb_idx.put(h.key(hg.name+"."+h.name), h)  # Register with alternative (full) name
            else:
                h = h[0]

            # Add the Hierarchy to the HierarchyGroup (if not)
            if h not in hg.hierarchies:
                hg.hierarchies.append(h)

            # Level
            level = item.get("level", None)
            if level:
                # Check if the level is defined. YES, get it; NO, create it
                for l in h.levels:
                    if strcmp(l.name, level):
                        level = l
                        break
                else:
                    level = HierarchyLevel(name=level, hierarchy=h)
                    h.levels.append(level)

            code = item.get("code", None)
            label = item.get("label", None)
            description = item.get("description", None)
            attributes = item.get("attributes", None)
            expression = item.get("expression", None)

            # Parent property (what really defines Hierarchies)
            parent_code = item.get("parent_code", None)
            if parent_code:
                ph = h  # Parent Hierarchy is the same as current hierarchy
                pcode = ph.codes.get(parent_code, None)
                if not pcode:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="Could not find code '"+parent_code+"' in hierarchy '"+ph.name+"'. Skipped.",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue
            else:
                pcode = None

            # ReferredHierarchy. If we are not defining a Code List, the base hierarchy has to be mentioned
            if not is_code_list:
                ref_hierarchy = item.get("referred_hierarchy", None)
                if not ref_hierarchy:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="For HCLs, defining ReferredHierarchy is mandatory",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue

                tmp = ref_hierarchy
                ref_hierarchy = glb_idx.get(Hierarchy.partial_key(name=ref_hierarchy))
                if len(ref_hierarchy) == 0:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="ReferredHierarchy '"+tmp+"' not defined previously",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue
                else:
                    ref_hierarchy = ref_hierarchy[0]

                ref_code = ref_hierarchy.codes.get(code, None)
                if not ref_code:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="Code '"+code+"' not found in referred hierarchy '"+ref_hierarchy.name+"'",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue

                # Ignore: LABEL, DESCRIPTION. Copy them from referred code
                label = ref_code.label
                description = ref_code.description
            else:
                ref_code = None

            c = h.codes.get(code, None)
            if c:
                issues.append(Issue(itype=IType.ERROR,
                                    description="Code '" + code + "' in hierarchy '" + h.name + "' redefined.",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                continue

            # Finally, create the HierarchyCode with all the gathered attributes, then weave it to other
            # (name, label=None, description=None, referred_node=None, parent=None, parent_weight=1.0, hierarchy=None)
            c = Taxon(name=code, hierarchy=h, level=level,
                      referred_taxon=ref_code, parent=pcode,
                      label=label, description=description,
                      attributes=attributes, expression=expression)
            # Add code to hierarchy
            h.codes[code] = c
            if not c.parent:
                h.roots_append(c)
            # Add code to level
            if level:
                level.codes.add(c)
            # Add child to parent code
            # (DONE BY THE CONSTRUCTOR!!)
            # if pcode:
            #     pcode.children_codes.append(c)

        return issues, None  # Issues, Output

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



# ONE TIME  -----------------------------------------------------------------------------
#
# pip install --upgrade setuptools wheel twine
#
# Create account:
# PyPI test: https://test.pypi.org/account/register/
# or PyPI  : https://pypi.org/account/register/
#
# EACH TIME -----------------------------------------------------------------------------
#
# Modify version code in "setup.py" (this file)
#
# Build (cd to directory where "setup.py" is) ###
# python3 setup.py sdist bdist_wheel
#
# Upload:
# PyPI test: twine upload --skip-existing --repository-url https://test.pypi.org/legacy/ dist/*
# or PyPI  : twine upload --skip-existing dist/*
#
# INSTALL   ------------------------------------------------------------------------------
#
# PyPI test: pip install --index-url https://test.pypi.org/simple/ --upgrade nexinfosys
# PyPI     : pip install --upgrade nexinfosys
# No PyPI  : pip install -e <local path where "setup.py" (this file) is located>
#      e.g.: pip install -e file:///home/rnebot/Dropbox/nis-backend
#
# EXECUTION EXAMPLE ("gunicorn" must be installed: "pip install gunicorn")
#
# pip install openpyxl==2.4.8 pandas==1.0.3
# gunicorn --workers=1 --log-level=debug --timeout=2000 --bind 0.0.0.0:8081 nexinfosys.restful_service.service_main:app
#
from os import path

from setuptools import setup
from pkg_resources import yield_lines
# from distutils.extension import Extension
from Cython.Build import cythonize
# from Cython.Distutils import build_ext

"""
python3 setup.py sdist bdist_wheel
twine upload --skip-existing dist/*

https://pypi.org/project/nexinfosys/
"""
package_name = 'nexinfosys'
version = '0.76'


def parse_requirements(strs):
    """Yield ``Requirement`` objects for each specification in `strs`

    `strs` must be a string, or a (possibly-nested) iterable thereof.
    """
    # create a steppable iterator, so we can handle \-continuations
    lines = iter(yield_lines(strs))

    ret = []
    for line in lines:
        # Drop comments -- a hash without a space may be in a URL.
        if ' #' in line:
            line = line[:line.find(' #')]
        # If there is a line continuation, drop it, and append the next line.
        if line.endswith('\\'):
            line = line[:-2].strip()
            try:
                line += next(lines)
            except StopIteration:
                return
        ret.append(line)

    return ret


with open('requirements-as-package.txt') as f:
    required = f.read().splitlines()

install_reqs = parse_requirements(required)
print(install_reqs)

# ext_modules = [
#     Extension("helper_accel", ["nexinfosys/common/helper_accel.pyx"]),
#     Extension("parser_spreadsheet_utils_accel", ["nexinfosys/command_generators/parser_spreadsheet_utils_accel.pyx"])
# ]

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name=package_name,
    version=version,
    install_requires=install_reqs,
    packages=['nexinfosys', 'nexinfosys.bin', 'nexinfosys.common', 'nexinfosys.models', 'nexinfosys.models.experiments',
              'nexinfosys.solving', 'nexinfosys.solving.graph', 'nexinfosys.ie_exports', 'nexinfosys.ie_imports',
              'nexinfosys.ie_imports.data_sources', 'nexinfosys.ie_imports.experimental', 'nexinfosys.authentication',
              'nexinfosys.model_services', 'nexinfosys.command_executors', 'nexinfosys.command_executors.misc',
              'nexinfosys.command_executors.solving', 'nexinfosys.command_executors.analysis',
              'nexinfosys.command_executors.version2', 'nexinfosys.command_executors.read_query',
              'nexinfosys.command_executors.external_data', 'nexinfosys.command_executors.specification',
              'nexinfosys.command_generators', 'nexinfosys.command_generators.spreadsheet_command_parsers',
              'nexinfosys.command_generators.spreadsheet_command_parsers.analysis',
              'nexinfosys.command_generators.spreadsheet_command_parsers.external_data',
              'nexinfosys.command_generators.spreadsheet_command_parsers.specification',
              # 'nexinfosys.magic_specific_integrations',
              'nexinfosys.command_generators.spreadsheet_command_parsers_v2',
              'nexinfosys.restful_service',
              'nexinfosys.restful_service.gunicorn',
              'nexinfosys.restful_service.mod_wsgi',
              'nexinfosys.restful_service.static.images',
              ],
    # See files to pack in "MANIFEST.in" file ("frontend" currently disabled)
    include_package_data=True,
    # cmdclass={'build_ext': build_ext},
    # ext_modules=cythonize(["nexinfosys/common/helper_accel.pyx", "nexinfosys/command_generators/parser_spreadsheet_utils_accel.pyx"], language_level="3"),
    url='https://github.com/MAGIC-nexus/nis-backend',
    license='BSD-3',
    author=['Rafael Nebot', 'Marco Galluzzi'],
    author_email='rnebot@itccanarias.org',
    entry_points={'console_scripts': ['nexinfosys=nexinfosys.bin.cli_script:main']},
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Formal and executable MuSIASEM multi-system Nexus models for Sustainable Development Analysis'
)



import json

from nexinfosys.command_generators import Issue, IssueLocation, parser_field_parsers, IType
from nexinfosys.command_generators.parser_ast_evaluators import dictionary_from_key_value_list
from nexinfosys.command_generators.parser_field_parsers import url_parser
from nexinfosys.common.helper import create_dictionary, load_dataset, prepare_dataframe_after_external_read, strcmp, \
    any_error_issue
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.models import CodeImmutable
from nexinfosys.models.musiasem_concepts import Hierarchy
from nexinfosys.models.musiasem_concepts_helper import convert_hierarchy_to_code_list
from nexinfosys.models.statistical_datasets import Dataset, Dimension, CodeList


class DatasetDefCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        def process_line(item):
            # Read variables
            dsd_dataset_name = item.get("dataset_name", None)
            dsd_dataset_data_location = item.get("dataset_data_location", None)
            dsd_concept_type = item.get("concept_type", None)
            if dsd_concept_type.lower() == "dataset":
                dsd_concept_type = None
            dsd_concept_name = item.get("concept_name", None)
            dsd_concept_data_type = item.get("concept_data_type", None)
            dsd_concept_domain = item.get("concept_domain", None)
            dsd_concept_description = item.get("concept_description", None)
            dsd_attributes = item.get("concept_attributes", None)
            if dsd_attributes:
                try:
                    attributes = dictionary_from_key_value_list(dsd_attributes, glb_idx)
                except Exception as e:
                    issues.append(Issue(itype=IType.ERROR,
                                        description=str(e),
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    return
            else:
                attributes = {}

            if dsd_dataset_name in ds_names:
                issues.append(Issue(itype=IType.ERROR,
                                    description="The dataset '"+dsd_dataset_name+"' has been already defined",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                return

            # Internal dataset definitions cache
            ds = current_ds.get(dsd_dataset_name, None)
            if True:  # Statistical dataset format
                if not ds:
                    ds = Dataset()
                    ds.code = dsd_dataset_name  # Name
                    ds.database = None
                    ds.attributes = {}
                    current_ds[dsd_dataset_name] = ds
                if not dsd_concept_type:
                    if ds.attributes.get("_location"):
                        issues.append(Issue(itype=IType.WARNING,
                                            description=f"Location of data for dataset {ds.code} previously declared. "
                                                        f"Former: {attributes.get('_location')}, "
                                                        f"Current: {dsd_dataset_data_location}",
                                            location=IssueLocation(sheet_name=name, row=r, column=None)))
                        attributes = ds.attributes
                    else:
                        attributes["_dataset_first_row"] = r
                    attributes["_location"] = dsd_dataset_data_location  # Location
                    ds.description = dsd_concept_description
                    ds.attributes = attributes  # Set attributes
                else:  # If concept_type is defined => add a concept
                    # Check if the concept name already appears --> Error
                    for d1 in ds.dimensions:
                        if strcmp(d1.code, dsd_concept_name):
                            issues.append(Issue(itype=IType.ERROR,
                                                description=f"Concept {dsd_concept_name} already declared for dataset {ds.code}",
                                                location=IssueLocation(sheet_name=name, row=r, column=None)))
                            break

                    d = Dimension()
                    d.dataset = ds
                    d.description = dsd_concept_description
                    d.code = dsd_concept_name
                    d.is_measure = False if dsd_concept_type.lower() == "dimension" else True
                    if not d.is_measure and dsd_concept_data_type.lower() == "time":
                        d.is_time = True
                    else:
                        d.is_time = False
                    if dsd_concept_type.lower() == "attribute":
                        attributes["_attribute"] = True
                    else:
                        attributes["_attribute"] = False
                    if dsd_concept_data_type.lower() == "category":
                        # TODO "hierarchies" variable really does not register hierarchies (see "hierarchy_command.py" or "hierarchy_categories_command.py", no insertion is made)
                        # h = hierarchies.get(dsd_concept_domain, None)
                        h = glb_idx.get(Hierarchy.partial_key(name=dsd_concept_domain))
                        if len(h) == 0:
                            issues.append(Issue(itype=IType.ERROR,
                                                description="Could not find hierarchy of Categories '" + dsd_concept_domain + "'",
                                                location=IssueLocation(sheet_name=name, row=r, column=None)))
                            return
                        elif len(h) > 1:
                            issues.append(Issue(itype=IType.ERROR,
                                                description="Found more than one instance of Categories '" + dsd_concept_domain + "'",
                                                location=IssueLocation(sheet_name=name, row=r, column=None)))
                            return
                        else:  # len(h) == 1
                            h = h[0]
                        d.hierarchy = h
                        # Reencode the Hierarchy as a CodeList
                        cl = convert_hierarchy_to_code_list(h)
                        d.code_list = cl

                    attributes["_datatype"] = dsd_concept_data_type
                    attributes["_domain"] = dsd_concept_domain
                    d.attributes = attributes

        # -------------------------------------------------------------------------------------------------------------
        issues = []
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        name = self._content["command_name"]

        # List of available dataset names. The newly defined datasets must not be in this list
        ds_names = [ds.name for ds in datasets]

        # List of available Category hierarchies
        hierarchies = create_dictionary()
        for h in hh:
            hierarchies[h.name] = hh

        # Datasets being defined in this Worksheet
        current_ds = create_dictionary()  # type: Dict[str, Dataset]

        # Process parsed information
        for line in self._content["items"]:
            r = line["_row"]
            # If the line contains a reference to a dataset or hierarchy, expand it
            # If not, process it directly
            is_expansion = False
            if is_expansion:
                pass
            else:
                process_line(line)

        # Any error?
        error = any_error_issue(issues)

        # Load the data for those datasets that are not local (data defined later in the same spreadsheet)
        for ds in current_ds.values():
            if "_location" not in ds.attributes:
                error = True
                issues.append(Issue(itype=IType.ERROR,
                                    description="Location of data not specified, for dataset '" + ds.code + "'",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
            else:
                loc = ds.attributes["_location"]
                ast = parser_field_parsers.string_to_ast(url_parser, loc)
                if ast["scheme"] != "data":
                    df = load_dataset(loc)
                    if df is None:
                        error = True
                        issues.append(Issue(itype=IType.ERROR,
                                            description=f"Could not obtain data for dataset '{ds.code}' at '{loc}'",
                                            location=IssueLocation(sheet_name=name, row=r, column=None)))
                    else:
                        iss = prepare_dataframe_after_external_read(ds, df, name)
                        issues.extend(iss)
                        # Everything ok? Store the dataframe!
                        if not any_error_issue(iss):
                            ds.data = df

        if not error:
            # If no error happened, add the new Datasets to the Datasets in the "global" state
            for ds in current_ds:
                r = current_ds[ds].attributes["_dataset_first_row"]
                df = current_ds[ds].data
                if df is not None:
                    # Loop over "ds" concepts.
                    # - "dimension" concepts of type "string" generate a CodeHierarchy
                    # - Check that the DataFrame contains ALL declared concepts. If not, generate issue
                    cid = create_dictionary(data={col: col for col in df.columns})
                    col_names = list(df.columns)
                    for c in current_ds[ds].dimensions:
                        if c.code in df.columns:
                            col_names[df.columns.get_loc(cid[c.code])] = c.code  # Rename column
                            dsd_concept_data_type = c.attributes["_datatype"]
                            if dsd_concept_data_type.lower() == "string" and not c.is_measure:  # Freely defined dimension
                                cl = df[cid[c.code]].unique().tolist()
                                # Filter "nan", with c == c (a Python idiom achieving this...)
                                cl = [c if c == c else "" for c in cl]
                                c.code_list = CodeList.construct(
                                    c.code, c.code, [""],
                                    codes=[CodeImmutable(c, c, "", []) for c in cl]
                                )
                        else:
                            issues.append(Issue(itype=IType.ERROR,
                                                description=f"Concept '{c.code}' not defined for '{ds}' in {loc}",
                                                location=IssueLocation(sheet_name=name, row=r, column=None)))
                    df.columns = col_names
                datasets[ds] = current_ds[ds]

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues



from __future__ import print_function

import io
import logging
import pickle
import os.path
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request

# If modifying these scopes, delete the file token.pickle.
from googleapiclient.errors import HttpError
from googleapiclient.http import MediaIoBaseDownload

SCOPES = ['https://www.googleapis.com/auth/drive.metadata.readonly',
          'https://www.googleapis.com/auth/drive.readonly']


def export_xlsx_to_memory(service, file_id):
    def process_request(request):
        fh = io.BytesIO()
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
            logging.debug("Download %d%%." % int(status.progress() * 100))
        return fh

    try:
        request = service.files().get_media(fileId=file_id)
        return process_request(request)
    except HttpError as e:
        try:
            request = service.files().export_media(fileId=file_id, mimeType="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet")
            return process_request(request)
        except HttpError as e:
            return None


def create_service(credentials_file, scopes, pickled_token_file):
    creds = None
    # The file token.pickle stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists(pickled_token_file):
        with open(pickled_token_file, 'rb') as token:
            creds = pickle.load(token)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(credentials_file, scopes)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(pickled_token_file, 'wb') as token:
            pickle.dump(creds, token)

    return build('drive', 'v3', credentials=creds)


def download_xlsx_file_id(credentials_path, token_path, file_id):
    service = create_service(credentials_path, SCOPES, token_path)
    return export_xlsx_to_memory(service, file_id)


def main():
    # How to get your "credentials.json" file:
    #
    # * Go to http://console.developers.google.com/apis/library
    # * Create project
    # * Go to "APIs y servicios" -> "Credenciales"
    # * Select "Crear credenciales" -> "Ayúdame a elegir"
    # * Select
    # ** ¿Qué API estás utilizando?: Google Drive API
    # ** ¿Desde dónde llamarás a la API?: Otra UI
    # ** ¿A qué tipo de datos accederás?: Datos de usuario


    # File ID. Go to any Google Calc or XLSX (open it converting to Google Calc first), then take the file ID from the
    # URL. It has 44 or 33 characters long.
    file_id = '--------------------------------------------'
    file_id = '---------------------------------'
    fh = download_xlsx_file_id('/home/rnebot/Dropbox/nis-backend-config/credentials.json', '/home/rnebot/Dropbox/nis-backend-config/token.pickle', file_id)
    if fh:
        with open("/home/rnebot/d.xlsx", 'wb') as out:
            out.write(fh.getvalue())
    else:
        print("Download failed!")


if __name__ == '__main__':
    main()



import json
import logging

from anytree import Node

from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.common.helper import obtain_dataset_metadata, strcmp, create_dictionary
from nexinfosys.models.musiasem_concepts import Mapping


def convert_code_list_to_hierarchy(cl, as_list=False):
    """
    Receives a list of codes. Codes are sorted lexicographically (to include numbers).

    Two types of coding schemes are supported by assuming that trailing zeros can be ignored to match parent -> child
    relations. The first is uniformly sized codes (those with trailing zeros). The second is growing length codes.

    Those with length less than others but common prefix are parents

    :param cl:
    :param as_list: if True, return a flat tree (all nodes are siblings, descending from a single root)
    :return:
    """

    def can_be_child(parent_candidate, child_candidate):
        # Strip zeros to the right, from parent_candidate, and
        # check if the child starts with the resulting substring
        return child_candidate.startswith(parent_candidate.rstrip("0"))

    root = Node("")
    path = [root]
    code_to_node = create_dictionary()
    for c in sorted(cl):
        if as_list:
            n = Node(c, path[-1])
        else:
            found = False
            while len(path) > 0 and not found:
                if can_be_child(path[-1].name, c):
                    found = True
                else:
                    path.pop()
            if c.rstrip("0") == path[-1].name:
                # Just modify (it may enter here only in the root node)
                path[-1].name = c
                n = path[-1]
            else:
                # Create node and append it to the active path
                n = Node(c, path[-1])
                path.append(n)
        code_to_node[c] = n  # Map the code to the node

    return root, code_to_node


def map_codelists(src, dst, corresp, dst_tree=False) -> (list, set):
    """
    Obtain map of two code lists
    If the source is a tree, children of a mapped node are assigned to the same mapped node
    The same source may be mapped more than once, to different nodes
    The codes from the source not mapped, are stored in "unmapped"

    :param src: source full code list
    :param dst: destination full code list
    :param corresp: list of tuples with the correspondence
    :param dst_tree: Is the dst code list a tree?
    :return: List of tuples (source code, target code), set of unmapped codes
    """

    def assign(n: str, v: str):
        """
        Assign a destination code name to a source code name
        If the source has children, assign the same destination to children, recursively

        :param n: Source code name
        :param v: Destination code name
        :return:
        """
        mapped.add(n, v)
        if n in unmapped:
            unmapped.remove(n)
        for c in cn_src[n].children:
            assign(c.name, v)

    unmapped = set(src)
    r_src, cn_src = convert_code_list_to_hierarchy(src, as_list=True)
    if dst_tree:
        r_dst, cn_dst = convert_code_list_to_hierarchy(dst)
    else:
        cn_dst = create_dictionary()
        for i in dst:
            cn_dst[i] = None  # Simply create the entry
    mapped = create_dictionary(multi_dict=True)  # MANY TO MANY
    for t in corresp:
        if t[0] in cn_src and t[1] in cn_dst:
            # Check that t[1] is a leaf node. If not, ERROR
            if isinstance(cn_dst[t[1]], Node) and len(cn_dst[t[1]].children) > 0:
                # TODO ERROR: the target destination code is not a leaf node
                pass
            else:
                # Node and its children (recursively) correspond to t[1]
                assign(t[0], t[1])

    for k in sorted(unmapped):
        logging.debug("Unmapped: " + k)
    # for k in sorted(r):
    #     print(k+" -> "+r[k])

    # Convert mapped to a list of tuples
    # Upper case
    mapped_lst = []
    for k in mapped:
        for i in mapped.getall(k):
            mapped_lst.append((k, i))

    return mapped_lst, unmapped


# Combinatory of lists (for many-to-many mappings)
#
# import itertools
# a = [[(0.5, "a"),(4, "b"),(7, "c")],[(1, "b")],[(0.2, "z"),(0.1, "y"),(0.7, "x")]]
# list(itertools.product(*a))
#

def fill_map_with_all_origin_categories(dim, map):
    # Check all codes exist
    mapped_codes = set([d["o"] for d in map])
    all_codes = set([c for c in dim.code_list])
    for c in all_codes - mapped_codes:  # Loop over "unmapped" origin codes
        # This sentence MODIFIES map, so it is not necessary to return it
        map.append({"o": c, "to": [{"d": None, "w": 1.0}]})  # Map to placeholder, with weight 1

    return map


class MappingCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        some_error = False
        issues = []
        # Read mapping parameters
        origin_dataset = self._content["origin_dataset"]
        origin_dimension = self._content["origin_dimension"]
        destination = self._content["destination"]
        # [{"o": "", "to": [{"d": "", "w": ""}]}]
        # [ {o: origin category, to: [{d: destination category, w: weight assigned to destination category}] } ]
        map = self._content["map"]
        # Obtain the origin dataset Metadata, obtain the code list
        dims, attrs, meas = obtain_dataset_metadata(origin_dataset)
        if origin_dimension not in dims:
            some_error = True
            issues.append((3, "The origin dimension '"+origin_dimension+"' does not exist in dataset '"+origin_dataset+"'"))
        else:
            dim = dims[origin_dimension]
            map = fill_map_with_all_origin_categories(dim, map)
            # # Check all codes exist
            # src_code_list = [c for c in dim.code_list]
            # dst_code_set = set()
            # many_to_one_list = []
            # for i in map:
            #     o = i["o"]
            #     for j in i["to"]:
            #         d = j["d"]
            #         dst_code_set.add(d)
            #         many_to_one_list.append((o, d))
            # hierarchical_code = True
            # if hierarchical_code:
            #     mapped, unmapped = map_codelists(src_code_list, list(dst_code_set), many_to_one_list)
            # else:
            #     # Literal. All codes on the left MUST exist
            #     mapped = many_to_one_list
            #     for i in mapped:
            #         o = i["o"]
            #         if o not in dim.code_list:
            #             some_error = True
            #             issues.append((3, "The origin category '" + o + "' does not exist in dataset dimension '" + origin_dataset + "." +origin_dimension + "'"))

        if some_error:  # Issues at this point are errors, return if there are any
            return issues, None

        # Create and store the mapping
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        from nexinfosys.ie_imports.data_source_manager import DataSourceManager
        source = DataSourceManager.obtain_dataset_source(origin_dataset, datasets)

        mappings[self._name] = Mapping(self._name, source, origin_dataset, origin_dimension, destination, map)

        # TODO If the categories to the left are not totally covered, what to do?
        # TODO - If a non-listed category appears, remove the line
        # TODO - If a non-listed category appears, leave the target column NA

        # TODO - If there are datasets matching the origin, JOIN

        return None, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues


from typing import Dict, Any

from nexinfosys.command_executors import BasicCommand
from nexinfosys.command_field_definitions import get_command_fields_from_class
from nexinfosys.models.musiasem_concepts import MatrixIndicator


class MatrixIndicatorsCommand(BasicCommand):
    def __init__(self, name: str):
        BasicCommand.__init__(self, name, get_command_fields_from_class(self.__class__))

    def _process_row(self, fields: Dict[str, Any], subrow=None) -> None:
        """
        Create and register MatrixIndicator object

        :param fields:
        """
        indicator = MatrixIndicator(
                fields["indicator_name"],
                fields.get("scope"),
                fields.get("processors_selector"),
                fields.get("interfaces_selector"),
                fields.get("indicators_selector"),
                fields.get("attributes_selector"),
                fields.get("description")
        )
        self._glb_idx.put(indicator.key(), indicator)



import json

from nexinfosys.command_generators.spreadsheet_command_parsers.specification import ref_prof
from nexinfosys.models.musiasem_concepts import Reference
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects


class ReferencesCommand(IExecutableCommand):
    """ It is a mere data container
        Depending on the type, the format can be controlled with a predefined schema
    """
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        Process each of the references, simply storing them as Reference objects
        """
        glb_idx, _, _, _, _ = get_case_study_registry_objects(state)

        issues = []

        # Receive a list of validated references
        # Store them as independent objects
        for ref in self._content["references"]:
            if "ref_id" not in ref:
                issues.append((3, "'ref_id' field not found: "+str(ref)))
            else:
                ref_id = ref["ref_id"]
                ref_type = ref["type"]
                existing = glb_idx.get(Reference.partial_key(ref_id, ref_type))
                if len(existing) == 1:
                    issues.append((2, "Overwriting reference '"+ref_id+"' of type '"+ref_type+"'"))
                elif len(existing) > 1:
                    issues.append((3, "The reference '"+ref_id+"' of type '"+ref_type+"' is defined more than one time ("+str(len(existing))+")"))

                reference = Reference(ref_id, ref_type, ref)
                glb_idx.put(reference.key(), reference)

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the content
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)
        return issues



from openpyxl.worksheet.worksheet import Worksheet

from nexinfosys import IssuesLabelContentTripleType, AreaTupleType
from nexinfosys.command_generators.spreadsheet_command_parsers.specification import profile_field_name_sets, ref_prof


def find_profile(col_names):
    # Convert in set and remove "ref_id"
    cn = set(col_names)
    cn.remove("ref_id")
    for prof_name, fields in profile_field_name_sets().items():
        if fields.issuperset(cn):
            for prof in ref_prof:
                if prof["type"] == prof_name:
                    return prof
            raise Exception("Should not be here!?")
    return None  # Free profile (or misspelled field name)


def validate(v, field):
    return True


def parse_references_command(sh: Worksheet, area: AreaTupleType, name: str = None) -> IssuesLabelContentTripleType:
    """
    Elaborate a list of dictionaries {key: value} which can be reused by other objects, referring them by a unique ref_id field

    For the definition of keys:values, two options:
     * If the column has a header, that would be the key
     * If the column does not have a header, both key and value can be specified in the cell, separated by "->" or ":" (which one?)

    :param sh: Input worksheet
    :param area: Tuple (top, bottom, left, right) representing the rectangular area of the input worksheet where the
    command is present
    :return: list of issues (issue_type, message), command label, command content
    """

    some_error = False
    issues = []

    references = []

    column_names = []
    for c in range(area[2], area[3]):
        value = sh.cell(row=area[0], column=c).value
        column_names.append(value.lower())
    if "ref_id" not in column_names:
        issues.append((3, "'ref_id' column is mandatory"))

    if some_error:
        return issues, None, []

    # Determine the type of reference contained in the worksheet
    profile = find_profile(column_names)
    if profile:
        type_ = profile["type"]
        col2field = {}
        for col in column_names:
            for f in profile["fields"]:
                if col == f.name:
                    col2field[col] = f
                    break
    else:
        type_ = "free_form"

    # Read each row
    for r in range(area[0]+1, area[1]):
        ref = dict(type=type_)  # Result dictionary
        for c in range(area[2], area[3]):
            value = sh.cell(row=area[0], column=c).value
            col = column_names[c-area[2]]
            if col == "ref_id":
                # Validate "ref_id" ¿syntax rules?
                ref["ref_id"] = value
            else:
                ref[col] = value
                field = col2field[col]
                if not validate(value, field):
                    issues.append((3, "Problem validating field '"+field.name+"' with value: "+value))

        references.append(ref)

    content = dict(references=references)

    return issues, None, content
